#ifndef DMTSERVER_HH
#define DMTSERVER_HH

#include "MonServer.hh"
#include "SigFlag.hh"
#include "Interval.hh"
#include "Time.hh"

/**  {\large {\bf Introduction}} \\
  *  The DMTServer class is a base class for dataless processes that produce 
  *  objects for the DMT viewer. The DMTServer class is designed to
  *  provide a similar environment whether running in the foreground 
  *  ( {\it i.e.} under root) or in the background. The functions
  *  provided by the environment are:
  *  \begin{itemize}
  *    \item Parse common command line arguments.
  *    \item Dispatch a display calculation method.
  *    \item Handle dmtviewer requests
  *    \item Handling attention interrupts.
  *    \item Clean up on termination.
  *  \end{itemize}
  *
  *  {\large {\bf Writing a DMTServer Monitor}} \\
  *  The strategy for writing a monitor program is to create a class 
  *  (#MyClass#) based on #DMTServer# including a constructor, destructor and
  *  #ProcessData# method. An attention handler may also be provided if
  *  desired. The operations to be performed by each of these methods
  *  are: \\ \\
  *  {\bf Constructor (MyClass(int argc, const char* argv[])) }
  *  \begin{itemize}
  *    \item Process user specific arguments
  *    \item Define parameters (see #TrigPar# class).
  *    \item Configure the process.
  *  \end{itemize}
  *
  *  {\bf Destructor (~MyClass(void)) }
  *  \begin{itemize}
  *    \item Save modified parameters
  *    \item Print results and statistics.
  *  \end{itemize}
  *
  *  {\bf Epoch processing (void ProcessData(void)) } \begin{itemize}
  *    \item Process data from one epoch.
  *    \item Keep statistics.
  *  \end{itemize}
  *
  *  {\bf Interrupt handler (void Attention(void)) }
  *  \begin{itemize}
  *    \item Monitor dependent interrupt processing.
  *  \end{itemize}
  *
  *  The #EXECSVR(myclass)# macro is used to generate the main
  *  function. Definition and implementation templates are available 
  *  from DatTemplate.hh and DatTemplate.cc.
  *
  *  The #EXEROOT(myclass)# macro is used to generate a main function
  *  with #root# graphics. Note that the "TROOT.h" and "TApplication.h"
  *  header files must be included in your monitor implementation file 
  *  if #EXEROOT# is used.
  *
  *  {\large {\bf Running the Monitor}} \\
  *  The base class constructor recognizes the following command line 
  *  arguments: 
  *  \begin{center} \begin{tabular}{ll}
  *  -debug <lvl>      & Set debug level to <lvl> \\
  *  -dt    <delta-t>  & Set reschedule time to <delta-t> \\
  *  -name  <serverid> & Set server name \\
  *  \end{tabular} \end{center}
  *
  *  @memo Data processing monitor base class.
  *  @author John G. Zweizig
  *  @version $id:$
  */
class DMTServer : public MonServer {
public:
    /**  The class constructor gets all the command line arguments in the 
      *  usual Unix/C argument list format. It sets up the environment 
      *  for processing and opens the frame input stream as directed by 
      *  the command line arguments.
      *  @memo Class constructor.
      *  @param argc Number of arguments specified.
      *  @param argv Pointers to arguments.
      */
  DMTServer(int argc, const char* argv[]);

    /**  Close the input stream and go away.
      *  @memo Class destructor.
      */
  virtual ~DMTServer(void);


    /**  #finish# stops processing by setting a flag that breaks the main 
      *  loop. Note that #finish# returns to the calling function (generally 
      *  the user's #ProcessData#) which must exit normally for the process 
      *  to terminate.
      *  @memo Stop data processing and go away cleanly.
      */
  void finish(void);

    /**  #Debug# returns the debug level as specified by the "-debug nnn" 
      *  argument. If "-debug" isn't specified in the command line the 
      *  debug level defaults to 0.
      *  @memo Get the debug level.
      */
  int Debug(void) const;

  Interval getReschedule(void) const;

  Time getWakeup(void) const;

    /**  Test for a DMTServer command line argument. All DMTServer command 
      *  line arguments take a value in the following argument position.
      *  @memo Test if string is a DMTServer command line argument.
      *  @return true if the specified string is a DMTServer argument.
      */
  bool isServerArg(const std::string& arg) const;

    /**  #MainLoop# redispatches #ProcessData# for each epoch. #MainLoop# 
      *  exits when a SIGINT or SIGTERM signal is caught, or when #finish# 
      *  is called by a subsidiary function (usually by #ProcessData#).This 
      *  function is provided by the  base class.
      *  @memo Loop over data epochs.
      */
  int MainLoop(void);
  void setReschedule(Interval dt);
  void setWakeup(const Time& t);
  void sleep_til(const Time& t);

    /**  #ProcessData# performs the monitor specific processing data from
      *  a single epoch. It must be defined as part of the derived monitor 
      *  class implementation.
      *  @memo Process one epoch (Supplied by user).
      */
  virtual void ProcessData(const Time& t) = 0;

    /**  #Attention# performs the monitor specific attention interrupt 
      *  processing. Its implementation is optional. Execution of 
      *  the attention handler is deferred until the processing of the 
      *  current epoch has completed. This means that there may be a 
      *  significant delay between the delivery of the SIGUSR1 signal and
      *  the invocation of the handler. The ROOT version uses the ROOT 
      *  event handler as a source of When used with ROOT, the default action is to
      *  terminate the monitor. When used in the background, it defaults
      *  to no action. The execution of the attention handler is 
      *  synchronized to the 
      *  @memo Attention interrupt handler (Supplied by user).
      */
  virtual void Attention(void);
private:
    bool     mActive;
    Interval mReschedule;
    Time     mWakeup;
    SigFlag mTerm;
    SigFlag mAttn;
    int     mDebug;
};

//
//   The EXECSVR macro generates the main function to execute a server
//   based on DMTServer. The argument is whatever name you have chosen
//   for your class.
//
#define EXECSVR(myclass) \
int main(int argc, const char *argv[]) { \
    myclass MyObject(argc,argv); \
    MyObject.MainLoop(); \
    return 0;}

//======================================  inline methods.
inline int
DMTServer::Debug(void) const {
    return mDebug;
}

inline Time
DMTServer::getWakeup(void) const {
    return mWakeup;
}

inline Interval
DMTServer::getReschedule(void) const {
    return mReschedule;
}

#endif   // !defined(DMTSERVER_HH)
