#ifndef GENERATOR_DAMPEDSINE_HH
#define GENERATOR_DAMPEDSINE_HH

#include "DataSource.hh"

namespace generator {

  /**  DampedSine is a damped sinusoid generator. The generated waveform is:
    *  \begin{verbatim}
           x(t)  = A * sin(omega*t + phi) * exp(-(t-to)*omega/Q )
       \end{verbatim}
    *  where the parameters are derived from the Generator Parameters
    *  as follows: A='A', omega=2*pi*'F' and phi='Phi'. The parameter 
    *  'Width' specifies the duration of the waveform to be generated 
    *  and is specified in units of Q/omega. At the default Width value of 
    *  6.0, the gaussian envelope has decreased to 1.25x10^-4 of its initial
    *  amplitude.
    *  @memo Damped sinusoid data source.
    *  @version 1.1; Last modified February 23, 2004
    *  @author John Zweizig
    */
  class DampedSine : public DataSource {
  public:

      /** Default constructor.
        */
      DampedSine(void);

      /**  Parameter constructor
        *  @param Amp  is the amplitude (A).
	*  @param Freq is the sinusoid frequency (F).
	*  @param Q    is the quality factor (Q).
	*  @param Phi  is the initial phase (Phi).
	*  @param wid  is the generation width.
	*/
      DampedSine(double Amp, double Freq, double Q, double Phi=0.0, 
		 double wid=6.0);

      /**  Destructor
        */
      ~DampedSine(void);

      /**  Construct an identical data source.
        */
      DampedSine* clone(void) const;

      /**  This function returns a literal string "DampedSine";
        *  @memo Get the data source type name.
        */
      const char* getDataType(void) const;

      /**  Returns the duration of generated data. This is calculated on an
        *  event-by-event basis as Q*Width/(2*pi*F).
        *  @memo Get the duration of data to be generated.
        *  @return Duraiton of generated data.
        */
      Interval getDuration(void) const;

      /**  Returns the start of generated data relative to the critical 
        *  time specified. For DampedSine this always returns zero.
        *  @memo Get the duration of data to be generated.
        *  @return Duration of generated data.
        */
      Interval getStartOffset(void) const;

  private:
      /**  Get generated data series for one event.
        *  @memo  Get generated data series.
	*/
      void getSeries(const Time& t0, Interval dT, int N, float* data);
  };

  //====================================  Inline functions.
  inline const char*
  DampedSine::getDataType(void) const {
      return "DampedSine";
  }

  inline Interval 
  DampedSine::getStartOffset(void) const {
      return Interval(0.0);
  }

} // namespace generator

#endif  // GENERATOR_DAMPEDSINE_HH
