#ifndef HTML_TABLE_HH
#define HTML_TABLE_HH

#include "html/compound.hh"
#include "html/Attrib.hh"
#include <vector>
#include <string>
#include <stdexcept>

namespace html {
  class align;

  /**  The HTML cell class is a compound object used inside a table.
    *  @memo Table cell class.
    *  @author J, Zweizig
    *  @version 1.0; Modified January 25, 2001
    */
  class cell : public compound {
  public:
    /**  Construct an empty cell.
      *  @memo Default constructor.
      */
    cell(void);

    /**  Construct a copy of an existing cell.
      *  @memo Copy constructor.
      */
    cell(const cell& x);

    /**  Destroy a cell.
      *  @memo Cell destructor
      */
    ~cell(void);

    /**  Create an exact copy of the cell. This implements the virtual
      *  html::object function.
      *  @return Pointer to a new cell object,`.
      */
    cell* clone(void) const;

    /**  Return the html tag associated with a cell in an html table.
      *  @memo return the Cell tag ("td").
      */
    const char* getTag(void) const {return "td";}

    /**  Set the alignment attribute for this cell.
      *  @memo Set the alignment attribute.
      */
    void setAlign(const std::string& a);

    /**  Add the binary save attribute with the specified value.
      *  @memo set the save attribute.
      *  @param on Initial value for the save attribute.
      */
    void setSave(bool on);
  };

  /**  The html table class stores a representation of an html table. The
    *  table contains an arbitrary number of rows and columns with each 
    *  entry containing a pointer to an html cell object. Each column has
    *  an optional header containing a title and attributes which are 
    *  inherited by all cells in the column.
    *  @memo Html Table object.
    *  @author J. Zweizig
    *  @version 1.1; Modified May 2, 2001
    */
  class table : public object {
  public:
    /**  Construct an empty table. 
      *  @memo Default constructor.
      */
    table(void);

    /**  Construct an empty html table and set the title.
      *  @memo Construct a table with a title
      *  @param initial title string
      */
    table(const std::string& title);

    /**  Construct an exact copy of the specified table.
      *  @memo Copy constructor.
      */
    table(const table& x);

    /**  Destroy the table.
      *  @memo Table destructor.
      */
    virtual ~table(void);

    /**  Create an exact copy of the table. This implements the virtual
      *  html::object function.
      *  @memo Clone a table.
      *  @return Pointer to a new table object.
      */
    table* clone(void) const;

    /**  Set the table contents to that of the argument table.
      *  @memo Table assignment operator.
      *  @param x Table whose contents are to be copied.
      */
    table& operator=(const table& x);

    /**  Write the table to an HTML document stream.
      *  @memo Write the table to the HTML document.
     */
    void write(writer& out) const;

    /**  Add a column to the table. all needed columns should be added
      *  to the table before adding rows or filling in cells.
      *  @memo Add a column.
      */
    int  addColumn(const std::string& title);

    /**  Add an empty row to the end of the table. The row number is 
      *  returned.
      */
    int  addRow(void);

    /**  Erase the specified row(s) from the table. All valid cells in 
      *  the deleted rows are deleted.
      *  @memo Erase one or more rows.
      *  @param Row Number of first row to be erased
      *  @param count Number of rows to be erased
      */
    void eraseRow(int row, int count=1) throw(std::runtime_error);

    /**  Insert floating point numeric data into the cell at the specified 
      *  row and column. If a cell exists at the specified location, its
      *  contents are replaced. Otherwise a new cell is created.
      */
    void insertData(int row, int col, double x, int wid=-1, int prec=-1) 
                    throw(std::runtime_error);

    /**  Insert integer data with an arbitrary radix into the cell at the 
      *  specified row and column. If a cell exists at the specified 
      *  location, its contents are replaced. Otherwise a new cell is 
      *  created.
      */
    void insertData(int row, int col, long x, int wid=-1, int radix=10) 
                    throw(std::runtime_error);

    /**  Insert text data into the cell at the specified row and column. 
      *  If a cell exists at the specified location, its contents are 
      *  replaced. Otherwise a new cell is created.
      */
    void insertData(int row, int col, const std::string& txt) 
                    throw(std::runtime_error);

    /**  Insert an arbitrary html object into the cell at the specified 
      *  row and column. If a cell exists at the specified location, its
      *  contents are replaced. Otherwise a new cell is created.
      */
    void insertData(int row, int col, const object& obj) 
                    throw(std::runtime_error);

    /**  Insert an empty row in front of the specified row number. The
      *  new row will have number #row# and all existing rows numbered
      *  #i >= row# will have their row number incremented. If #row# 
      *  is not valid (#row < 0# or #row > nRow#) a #runtime_error#
      *  exception will be thrown after creating a new row at the 
      *  end of the table.
      */
    void insertRow(int row) throw(std::runtime_error);

    /**  Get the current number of rows in the table.
      *  @memo Get the number of rows.
      */
    int getNRow(void) const;

    /**  Get the number of columns defined in the table.
      *  @memo Get the number of columns.
      */
    int getNColumn(void) const;

    /**  Get a reference to the header for the specified column.
      *  @memo Get a column header reference.
      */
    cell& refHeader(int col) throw(std::runtime_error);

    /**  Get a constant reference to the header for the specified column.
      *  @memo Get a column header reference.
      */
    const cell& refHeader(int col) const throw(std::runtime_error);

    /**  Get a reference to the cell at a specified location within the 
      *  table.
      *  @memo Get a specified cell reference.
      */
    cell& refCell(int row, int col) throw(std::runtime_error);

    /**  Get a constant reference to the cell at a specified location 
      *  within the table.
      *  @memo Get a specified cell reference.
      */
    const cell& refCell(int row, int col) const throw(std::runtime_error);

    /**  Set the table border attribute. If the border attribute is set, 
      *  a border is drawn around each cell of the table.
      *  @memo Enable/disable cell borders.
      */
    void setBorder(bool On=true);

  private:
    /**  Internal function to return the absolute index for the cell at
      *  the given row and column position.
      */
    int inxrc(int row, int col) const throw(std::runtime_error);

  private:
    std::string mTitle;
    int mRow;
    typedef std::vector<cell> ColHdr;
    ColHdr    mCTitle;
    attList   mAttrib;
    typedef std::vector<cell*> TableData;
    TableData mData;
  };

}

//======================================  Inline functions
inline int 
html::table::getNColumn(void) const {
    return mCTitle.size();
}

inline int 
html::table::getNRow(void) const {
    return mRow;
}

inline int
html::table::inxrc(int row, int col) const throw(std::runtime_error) {
    if (row >= mRow || row < 0) throw std::runtime_error("Invalid Row.");
    int nCol = getNColumn();
    if (col >= nCol || col < 0) throw std::runtime_error("Invalid Column.");
    return row*nCol + col;
}

inline const html::cell& 
html::table::refCell(int row, int col) const throw(std::runtime_error) {
    return *mData[inxrc(row, col)];
}

inline html::cell& 
html::table::refCell(int row, int col) throw(std::runtime_error) {
    return *mData[inxrc(row, col)];
}

inline const html::cell& 
html::table::refHeader(int col) const throw(std::runtime_error) {
    if (col<0 || col>=getNColumn()) throw std::runtime_error("Invalid Column.");
    return mCTitle[col];
}

inline html::cell& 
html::table::refHeader(int col) throw(std::runtime_error) {
    if (col<0 || col>=getNColumn()) throw std::runtime_error("Invalid Column.");
    return mCTitle[col];
}

#endif  // HTML_TABLE_HH
