#ifndef XSIL_EVENTTABLE_HH
#define XSIL_EVENTTABLE_HH

#include <time.h>
#include <string>
#include <list>
#include <iosfwd>

namespace xsil {

  class XSIL;
  class table;
  class Stream;


  /** Produce an XML document describing a trigger database table.
    * Each row starts on a new output line in the XML document.
    */
  class EventTable {
    struct eventcolumn {
      eventcolumn(void);
      eventcolumn(const eventcolumn& x);
      ~eventcolumn(void);
      std::string data_Type;
      const void* data_Addr;
    };

    typedef std::list<eventcolumn> ColumnList;
    typedef ColumnList::const_iterator col_iter;

  public:
    EventTable(void);
    EventTable(const char* Title);
    ~EventTable(void);

    /**  Set up the table internally. Specify the name and table type
      *  attributes.
      *  @memo make a table.
      */
    void makeTable(const char* Name, const char* Type);

    /**  Specify a table column. The column name and type are included.
      *  an address field specifies a static address from which the data 
      *  will be copied. The following types are currently supported:
      *  \begin{tabular}{|l|l|l}
      *  Type & c-pointer \\
      *  STRING & const char* \\
      *  INT4 & const int* \\
      *  REAL4 & const float* \\
      *  REAL8 & const double* \end{tabular}
      *  A '*' is added to the stream to indicate an invalid data type,
      *  and "00" is added to indicate a null pointer specified.
      *  @memo Add a column to the table.
      *  @param Name Column name attribute field.
      *  @param Type Column data type.
      *  @param Addr Pointer to the data field.
      */
    void addColumn(const char* Name, const char* Type, const void* Addr=0);

    /**  Copy all specified data fields (one for each table column) into the 
      *  data stream.
      *  @memo Copy an event into the data stream.
      */
    void putEvent(void);

    /**  Write the XML document to the specified output stream. The table
      *  data stream is cleared after the data have been written.
      *  @memo Write the XML document.
      *  @param out Output stream to which the data will be written.
      */
    void putTable(std::ostream& out);

    /**  Delete the contens of the table data stream.
      *  @memo Reset the data stream.
      */
    void resetStream(void);

  private:
    std::string mTitle;
    ColumnList mCList;
    XSIL*      mDoc;
    table*     mTable;
  };

}  // namespace xsil

inline
xsil::EventTable::eventcolumn::eventcolumn(void) {}

inline
xsil::EventTable::eventcolumn::eventcolumn(const eventcolumn& x) {
  data_Type = x.data_Type;
  data_Addr = x.data_Addr;
}

inline
xsil::EventTable::eventcolumn::~eventcolumn(void) {}


#endif  //  XSIL_EVENTTABLE_HH

