//    -*- C++ -*-
//
//    File ServoMon.cc
//    Trigger on servo instability
//    $Id: ServoMon.cc,v 1.25 2004/06/16 10:12:42 jzweizig Exp $

#include "ServoMon.hh"

#ifndef __CINT__

#include <iostream>
#include <iomanip>
#include <fstream>
#include <string>
#include <stdexcept>
#include <vector>
#include <stdlib.h>
#include <math.h>
#include "TSeries.hh"
#include "FSeries.hh"
#include "Dacc.hh"
#include "DatEnv.hh"
#include "TrigRslt.hh"

// The next three lines must appear, and in this sequence
#define  PIDCVSHDR "$Header: /ldcg_server/common/repository_gds/gds/Monitors/ServoMon/ServoMon.cc,v 1.25 2004/06/16 10:12:42 jzweizig Exp $"
#define  PIDTITLE  "Watch for servo instabilities (MC_F, DARM_CTRL only)"
#include "ProcIdent.hh"



//
// Generate the main routine.
//
EXECDAT(ServoMon) ;
#endif // !def(__CINT__)

using namespace std;

template <class T> static double sum(const std::vector<T> &data)
{
    double retval = 0.;

    typename std::vector<T>::const_iterator iter = data.begin();
    for (; iter != data.end(); ++iter)
        retval += (double)*iter;

    return retval;
}


template <class T> static double mean(const std::vector<T> &data)
{
    return (sum(data) / (double)(data.size()));
}

template <class T> static double variance(const std::vector<T> &data)
{
    double avg = mean(data);
    std::vector<T> delta = data;
    typename std::vector<T>::iterator iter = delta.begin();
    for (; iter != delta.end(); ++iter)
        *iter -= avg;
    return (sum(delta) / (double)(data.size() - 1));
}

template <class T> static double stdDev(const std::vector<T> &data)
{
    return sqrt(variance(data));
}

ServoMon::ServoMon(int argc, const char *argv[]) 
    : DatEnv(argc, argv),
      mName("ServoMon_"),
      MaxFrame(999999),
      mFrameCount(0),
      mMCFLastProb(0),
      mD0LastProb(0),
      mD0pLastProb(0),
      mD1LastProb(0),
      mD2LastProb(0),
      mOsclist(mIn, 0),
      mDebug(0),
      mProblemPstr(":MC_F_high_freq_problem"),
      mMCFstr(":MC_F_high_freq_power_3kHz-7kHz_12hr"),
      mD0str(":DARM_CTRL_340-345Hz_power_12hr"),
      mD0plusstr(":DARM_CTRL_345-355Hz_power_12hr"),
      mD1str(":DARM_CTRL_680-690_power_12hr"),
      mD2str(":DARM_CTRL_400-450Hz_power_12hr"),
      mMCFProblemPWindow(-1,mMaxLength),
      mPowerWindowMCF(-1,mMaxLength),
      mPowerWindowD0(-1,mMaxLength),
      mPowerWindowD0plus(-1,mMaxLength),
      mPowerWindowD1(-1,mMaxLength),
      mPowerWindowD2(-1,mMaxLength),
      mTrigType("Servo instabilities"),
      mTrigSubtypeMCF("MC_3kHz-7kHz"),
      mTrigSubtypeD0("DARM_340Hz-345Hz"),
      mTrigSubtypeD0plus("DARM_345Hz-355Hz"),
      mTrigSubtypeD1("DARM_680Hz-690Hz"),
      mTrigSubtypeD2("DARM_400Hz-450Hz"),
      mMCFcounter(0),
      mD0counter(0),
      mD0pCounter(0),
      mD1counter(0),
      mD2counter(0),
      mTstampFormat("%Y-%m-%d %H:%N:%S %Z"),
      mTotalRunTime(0,0),
      mMCFTriggers(-1,4320),
      mD0Triggers(-1,4320),
      mD0pTriggers(-1,4320),
      mD1Triggers(-1,4320),
      mD2Triggers(-1,4320)
{
    if (argc < 3) {
        std::string eObj = "Must specify IFO name (H1, H2, or L1)";
        cerr << eObj << endl;
        cerr << "Usage: ServoMon -ifo ifoname [-debug n] " << endl;
        throw runtime_error(eObj);
    }

    cerr.precision(12);

    bool ifoset_p = false;

    for (int i = 0; i < argc; ++i) {
        if (strcmp(argv[i], "-debug") == 0) {
            ++i;
            mDebug = atoi(argv[i]);
        }

        if (strcmp(argv[i], "-ifo") == 0) {
            ++i;

            if (i > argc-1 || argv[i] == NULL) {
                std::string eObj = "Must specify IFO name (H1, H2, or L1)";
                cerr << eObj << endl;
                throw runtime_error(eObj);
            }

            mIFOname = argv[i];
            ifoset_p = true;
            
            if (mIFOname != "H1" && mIFOname != "H2" && mIFOname != "L1") {
                std::string eObj = "Must specify IFO name (H1, H2, or L1)";
                cerr << "Must specify IFO name (H1, H2, or L1)" << endl;
                throw runtime_error(eObj);
            }
        }
    }

    if (ifoset_p == false) {
        std::string eObj = "Must specify IFO name (H1, H2, or L1)";
        cerr << eObj<< endl;
        throw runtime_error(eObj);
    }

    // Set monitor name
    mName += mIFOname;

    // Set monitor server name (for serving data to dmtviewer)
    if (mDebug == 0) {
        MonServer::setServerName(mName.c_str());
    } else {
        mName += "_DEBUG";
        MonServer::setServerName(mName.c_str());
    }

    // Set osclist debug level
    mOsclist.setDebug(mDebug);

    // Write to $DMTHTMLOUT (set by ProcessManager) or
    // $GDSAPACHE/monitor_reports/mName 
    char *htmldir = getenv("DMTHTMLOUT");
    if (!htmldir) {
        cerr << mName << ": warning -- environment variable DMTHTMLOUT "
             << "not set. Trying GDSAPACHE." << endl;
        
        htmldir = getenv("GDSAPACHE");
        if (!htmldir) {
            cerr << mName << ": warning -- environment variable GDSAPACHE "
                 << "not set. Writing log and duty cycle files to "
                 << "current directory" << endl;
        } else {
            cerr << mName << ": Writing log and duty cycle files to "
                 << "$GDSAPACHE/monitor_reports/" << mName << "/"
                 << endl;

            mHtmlDir  = htmldir;
            mHtmlDir += "/monitor_reports/" + mName + "/";
        }
    } else {
        cerr << mName << ": Writing log and duty cycle files to "
             << htmldir << endl;

        mHtmlDir = htmldir;
        mHtmlDir += "/";
    }

    mLogfileName   = mHtmlDir + "ServoMon_Log.txt";
    mIndexPageName = mHtmlDir + "index.html";
    mDispfileName  = mHtmlDir + "ServoMon_Status.html";
    std::string urldir;
    if (mIFOname == "H1")
        urldir = "http://blue.ligo-wa.caltech.edu/gds/Alarms/" + mName;
    else
        urldir = "http://london.ligo-la.caltech.edu/gds/Alarms/" + mName;
    //char *gdsapache = getenv("GDSAPACHE");
    //std::string urldir = gdsapache;
    //urldir += "dmt/Alarms/";
    //urldir += mName + "_";
    mMCF_alarmDesc = urldir + "_MCF_alarmDesc.html";
    mD0_alarmDesc = urldir + "_D0_alarmDesc.html";
    mD0p_alarmDesc = urldir + "_D0p_alarmDesc.html";
    mD1_alarmDesc = urldir + "_D1_alarmDesc.html";
    mD2_alarmDesc = urldir + "_D2_alarmDesc.html";

    if (mDebug > 0) {
        std::cerr << mName << ": mLogfileName = " << mLogfileName << std::endl;
        std::cerr << mName << ": mIndexPageName = " << mIndexPageName << std::endl;
        std::cerr << mName << ": mDispfileName = " << mDispfileName << std::endl;
        std::cerr << mName << ": mMCF_alarmDesc = " << mMCF_alarmDesc << std::endl;
    }


    writeIndexPage();

    // Open log with append mode
    mLogfile.open(mLogfileName.c_str(), ios::app);

    if (!mLogfile) {
        cerr << mName << ": could not open logfile " << mLogfileName << endl;
        cerr << "          reverting to stdout" << endl;
        mLog = &std::cout;
    } else {
        mLog = &mLogfile;
    }


    // Set alarm client name
    mAlarmClient.setName(mName.c_str());

    // Set alarm data and handles
    Interval timeout(10.0);
    int      severity(1);  // it is not specified in the AlarmData
                           // documentation how severities are ordered
    AlarmData tmp_ad1(mName, "MCF_hi_f_noise", timeout, severity,
                      mMCF_alarmDesc);
    AlarmData tmp_ad2(mName, "DARM_CTRL_340_345Hz_noise", timeout, severity,
                      mD0_alarmDesc);
    AlarmData tmp_ad3(mName, "DARM_CTRL_345_350Hz_noise", timeout, severity,
                      mD0p_alarmDesc);
    AlarmData tmp_ad4(mName, "DARM_CTRL_680_690Hz_noise", timeout, severity,
                      mD1_alarmDesc);
    AlarmData tmp_ad5(mName, "DARM_CTRL_400_450Hz_noise", timeout, severity,
                      mD2_alarmDesc);

    mMCF_ad = tmp_ad1;
    mD0_ad  = tmp_ad2;
    mD0p_ad = tmp_ad3;
    mD1_ad  = tmp_ad4;
    mD2_ad  = tmp_ad5;

    // return code for alarm settings
    lmsg::error_t rc;
    rc = mAlarmClient.defineAlarm(mMCF_ad);
    if (rc != 0) {
        // failure
        *mLog << mName << ": error defining alarm: MCF problem: return code "
              << rc << std::endl;
    }

    rc = mAlarmClient.defineAlarm(mD0_ad);
    if (rc != 0) {
        // failure
        *mLog << mName << ": error defining alarm: DARM_CTRL violin fundamental problem:"
              << " return code " << rc << std::endl;
    }
    
    rc = mAlarmClient.defineAlarm(mD0p_ad);
    if (rc != 0) {
        // failure
        *mLog << mName << ": error defining alarm: DARM_CTRL 345Hz-350Hz  problem:"
              << " return code " << rc << std::endl;
    }
        
    rc = mAlarmClient.defineAlarm(mD1_ad);
    if (rc != 0) {
        // failure
        *mLog << mName << ": error defining alarm: DARM_CTRL violin 1st harmonic  problem:"
              << " return code " << rc << std::endl;
    }
    
    rc = mAlarmClient.defineAlarm(mD2_ad);
    if (rc != 0) {
        // failure
        *mLog << mName << ": error defining alarm: DARM_CTRL 400Hz-450Hz problem:"
              << " return code " << rc << std::endl;
    }    

    Time now = mIn.getFillTime();
    *mLog << "Creating ServoMon object at time " << now.getS() << endl;

    // Set the time stride
    getDacc().setStride(mStride);

    setTSWindowStuff();

    // once the savefile names of the TSWindows are set
    // we can restore them
    restoreAllTSWindows();

    // Null out timestamps
    mMCFtimestamp[0] = '\0';
    mD0timestamp[0] = '\0';
    mD0pTimestamp[0] = '\0';
    mD1timestamp[0] = '\0';
    mD2timestamp[0] = '\0';

    // Set various input and output streams
    char *ifocstr = getenv("DMTIFOS");

    // Read in configuration file
    if (!ifocstr) {
        std::string eObj =
            "ServoMon: Don't know if we're at Hanford or Livingston. ";
        eObj += "Set the environment variable DMTIFOS and restart.";
        cerr << eObj
             << endl;
        throw runtime_error(eObj);
    } else {
        std::string ifos(ifocstr);
        // Make sure requested IFO is available at this site
        if (ifos.find(mIFOname) == std::string::npos) {
            std::string eObj = "Specified IFO (";
            eObj += mIFOname;
            eObj += ") is not available at this site";
            cerr << eObj << endl;
            throw runtime_error(eObj);
        } else {
            if (mDebug > 0) {
                // read config file from current directory
                cerr << "FOOBAR" << endl;
                mOsclist.readConfig("ServoMon.conf");
            } else {
                // read central config file
                cerr << "CUBAAN" << endl;
                mOsclist.readConfig("/export/home/ops/pars/ServoMon.conf");
            }
        }
    }

    // Dump all conditions that we don't care about
    if (mDebug > 0)
        cerr << mName << ": dumping unneeded IFOs" << endl;

    if (mDebug > 1)
        mOsclist.chanUtilInfo();

    mOsclist.ignoreAllExcept(mIFOname);
    
    cerr << mName << ": Done dumping" << endl;
    cerr << mName << ": DEBUG LEVEL = " << mDebug << endl;
    

    if (mDebug > 0) {
        mOsclist.chanUtilInfo();
        mOsclist.listOSCs();
        std::cerr << std::endl;
        std::cerr << mName << ": mOsclist.bucket_count() = "
                  << mOsclist.bucket_count() << std::endl;
    }
    
    // Once config file is read, query OSCs for channel names so we
    // know the IFO name with which to tag our triggers
    std::string ifoname = mOsclist[mIFOname + ":MC_F_peaking"]->channel();

    if (mIFOname != ifoname.substr(0,2)) {
        std::string eObj = "IFO name specified on command line (";
        eObj += mIFOname;
        eObj += ") does not match IFO name in config file (";
        eObj += ifoname;
        eObj += ")";
        cerr << eObj << endl;
        throw runtime_error(eObj);
    }

    if (mDebug > 0) {
        *mLog << "mIFOname = " << mIFOname << endl;
        *mLog << "about to serveData for DMTviewer" << endl;
    }

    // Serve data to DMTviewer
    serveData(mMCFProblemPWindow.name().c_str(), &mMCFProblemTS);
    serveData(mPowerWindowMCF.name().c_str(), &mPowerTSMCF);
    serveData(mPowerWindowD0.name().c_str(),  &mPowerTSD0);
    serveData(mPowerWindowD0plus.name().c_str(),  &mPowerTSD0plus);
    serveData(mPowerWindowD1.name().c_str(),  &mPowerTSD1);
    serveData(mPowerWindowD2.name().c_str(),  &mPowerTSD2);


    if (mDebug > 0) {
        *mLog << "Setting up Trends" << endl;
    }

    //
    // Set up trending
    //

    // trend channel names
    mTrendChanNames["MCF"] = mIFOname + ":IOO-MC_F_SM_3k-7k";
    mTrendChanNames["D0"]  = mIFOname + ":LSC-DARM_CTRL_SM_340-345";
    mTrendChanNames["D0p"] = mIFOname + ":LSC-DARM_CTRL_SM_345-355";
    mTrendChanNames["D1"]  = mIFOname + ":LSC-DARM_CTRL_SM_680-690";
    mTrendChanNames["D2"]  = mIFOname + ":LSC-DARM_CTRL_SM_400-450";

    // Trend index file is $DMTRENDOUT/channel.cfg
    char *tdir = getenv("DMTRENDOUT");
    std::string trendDir;
    if (!tdir) {
        trendDir = "/export/DMTtrends/";
        trendDir += mName;
        trendDir += "/";
    } else {
        trendDir = tdir;
    }

    mTrend.addChannel(mTrendChanNames["MCF"].c_str());
    mTrend.addChannel(mTrendChanNames["D0"].c_str());
    mTrend.addChannel(mTrendChanNames["D0p"].c_str());
    mTrend.addChannel(mTrendChanNames["D1"].c_str());
    mTrend.addChannel(mTrendChanNames["D2"].c_str());

    char indexfile[512];
    sprintf(indexfile, "%s/channel.cfg", trendDir.c_str());

    // trend frame
    mTrend.setIFO(mIFOname.c_str());
    mTrend.setName(mName.c_str());
    mTrend.setType(Trend::kMinute);
    mTrend.setFrameLen(60);  // 60 min == 1 hr per frame
    mTrend.setFrameCount(1); // 1 frame per file
    mTrend.setAutoUpdate(false);
    mTrend.writeIndex(indexfile);

    if (mDebug > 0) {
        *mLog << "Done setting up trends.  Serving trends to DMTviewer..."
              << endl;
    }

    // serve up trends to the DMTviewer
    const std::string ntag = "_min_trend_24hr";
    std::string dmtviewerName = mTrendChanNames["MCF"] + ntag;

    if (mDebug > 0) {
        *mLog << "    ntag = " << ntag << endl
              << "    dmtviewerName = " << dmtviewerName << endl;
    }

    if (mDebug > 0) {
        *mLog << "    ntag = " << ntag << endl
              << "    dmtviewerName = " << dmtviewerName << endl;
    }
    serveData(dmtviewerName.c_str(),
              &(mTrend.find(mTrendChanNames["MCF"].c_str()).refAvgSeries()),
              "ServoMon: IOO-MC_F power 3kHz-7kHz");

    if (mDebug > 0) {
        *mLog << "... done MCF trend" << endl;
    }

    dmtviewerName = mTrendChanNames["D0"] + ntag;
    serveData(dmtviewerName.c_str(),
              &(mTrend.find(mTrendChanNames["D0"].c_str()).refAvgSeries()),
              "ServoMon: LSC-DARM_CTRL power 340Hz-345Hz");

    if (mDebug > 0) {
        *mLog << "... done D0 trend" << endl;
    }

    dmtviewerName = mTrendChanNames["D0p"] + ntag;
    serveData(dmtviewerName.c_str(),
              &(mTrend.find(mTrendChanNames["D0p"].c_str()).refAvgSeries()),
              "ServoMon: LSC-DARM_CTRL power 345Hz-355Hz");

    if (mDebug > 0) {
        *mLog << "... done D0+ trend" << endl;
    }

    dmtviewerName = mTrendChanNames["D1"] + ntag;
    serveData(dmtviewerName.c_str(),
              &(mTrend.find(mTrendChanNames["D1"].c_str()).refAvgSeries()),
              "ServoMon: LSC-DARM_CTRL power 680Hz-690Hz");

    if (mDebug > 0) {
        *mLog << "... done D1 trend" << endl;
    }

    dmtviewerName = mTrendChanNames["D2"] + ntag;
    serveData(dmtviewerName.c_str(),
              &(mTrend.find(mTrendChanNames["D2"].c_str()).refAvgSeries()),
              "ServoMon: LSC-DARM_CTRL power 400Hz-450Hz");

    if (mDebug > 0) {
        *mLog << "... done D2 trend" << endl;
    }
}

//
// Destructor
//
ServoMon::~ServoMon() 
{
    saveAllTSWindows();
    *mLog << "ServoMon is finished" << endl << endl;
    *mLog << "##############################" << endl << endl;
    mLogfile.close();
}

//
// appendToLog()
//
void ServoMon::appendToLog(Time now)
{
    *mLog << mName << ": MC_F 3000-7000 power = ";
    sprintf(mOutputStr, "%12.8e",
            mOsclist[mIFOname + ":MC_F_peaking"]->watchedQuantity());
    *mLog << mOutputStr << " at time " << now.getS() << endl;
    
    *mLog << mName << ": DARM_CTRL 340-345 power = ";
    sprintf(mOutputStr, "%12.8e",
            mOsclist[mIFOname + ":DARM_CTRL_violin_0"]->watchedQuantity());
    *mLog << mOutputStr << " at time " << now.getS() << endl;
    
    *mLog << mName << ": DARM_CTRL 345-355 power = ";
    sprintf(mOutputStr, "%12.8e",
            mOsclist[mIFOname + ":DARM_CTRL_violin_0_plus"]->watchedQuantity());
    *mLog << mOutputStr << " at time " << now.getS() << endl;

    *mLog << mName << ": DARM_CTRL 400-455 power = ";
    sprintf(mOutputStr, "%12.8e",
            mOsclist[mIFOname + ":DARM_CTRL_400_450"]->watchedQuantity());
    *mLog << mOutputStr << " at time " << now.getS() << endl;
    
    *mLog << mName << ": DARM_CTRL 680-690 power = ";
    sprintf(mOutputStr, "%12.8e",
            mOsclist[mIFOname + ":DARM_CTRL_violin_1"]->watchedQuantity());
    *mLog << mOutputStr << " at time " << now.getS() << endl;
} // END: appendToLog()


//
// updatePowerWindows()
//
void ServoMon::updatePowerWindows(Time now)
{

    if (mOsclist[mIFOname + ":Both_arms_locked"]->satisfied()) {
        mPowerWindowMCF.append(
            mOsclist[mIFOname + ":MC_F_peaking"]->watchedQuantity(),
            now);

        mPowerWindowD0.append(
            mOsclist[mIFOname + ":DARM_CTRL_violin_0"]->watchedQuantity(),
            now);

        mPowerWindowD0plus.append(
            mOsclist[mIFOname + ":DARM_CTRL_violin_0_plus"]->watchedQuantity(),
            now);

        mPowerWindowD1.append(
            mOsclist[mIFOname + ":DARM_CTRL_violin_1"]->watchedQuantity(),
            now);

        mPowerWindowD2.append(
            mOsclist[mIFOname + ":DARM_CTRL_400_450"]->watchedQuantity(),
            now);
    } else {
        mPowerWindowMCF.append(0., now);

        mPowerWindowD0.append(0., now);

        mPowerWindowD0plus.append(0., now);

        mPowerWindowD1.append(0., now);

        mPowerWindowD2.append(0., now);
    }
    

    if (mDebug > 3) {
        cerr << mName << ": mPowerWindowD0plus.length() = "
             << mPowerWindowD0plus.length() << endl; 
        cerr << mName << ": mPowerWindowD0plus.name()   = "
             << mPowerWindowD0plus.name() << endl;

        std::vector<double> dvec;
        mPowerWindowD0plus.data(dvec);
        cerr << mName << ": dvec.size() = " << dvec.size() << endl;
        cerr << mName << ": dvec[1] = " << dvec[1]
             << "; dvec[1024] = " << dvec[1024]
             << "; dvec[mMaxLength-1] = " << dvec[mMaxLength-1] << endl;

        cerr << mName << ": mPowerTSD0plus.getNSample() = "
             << mPowerTSD0plus.getNSample() << endl;
    }

} // END: updatePowerWindows()


//
// MCFproblemActions(Time now)
//
void ServoMon::MCFproblemActions(Time now)
{
    if (mDebug > 0) {
        *mLog << mName << ": Entered MCFproblemActions" << std::endl;

        lmsg::error_t rc2 = mAlarmClient.setAlarm(mMCF_ad, mMCF_ah);
        if (rc2)
            *mLog << mName << ": Error setting alarm for MCF: error code "
                  << rc2 << std::endl;
        else
            *mLog << mName << ": Alarm MCF successfully set" << std::endl;
    }

    // increment counter
    ++mMCFcounter;

    // set timestamp
    TimeStr(now, mMCFtimestamp, mTstampFormat.c_str());
    
    // Set up trigger
    trig::TrigRslt tout(mTrigType.c_str(),mTrigSubtypeMCF.c_str(),0);
    tout.setIfos(mIFOname.c_str());
    tout.setTime(now);

    if (mDebug > 0) {
        *mLog << mName << ": Debug trigger of type '"
              << mTrigType << "', subtype '"
              << mTrigSubtypeMCF << "'" << endl;
        cerr << mName << ": Debug trigger of type '"
             << mTrigType << "', subtype '"
             << mTrigSubtypeMCF << "'" << endl;
    } else {
        // send trigger to metaDB and raise alarm
        tout.setDisposition(trig::d_metaDB + trig::d_alarm);
            
        *mLog << mName << ": Sending trigger of type '"
              << mTrigType << "', subtype '"
              << mTrigSubtypeMCF << "'" << endl;
        
        lmsg::error_type rc = sendTrigger(tout);
        if (rc)
            *mLog << mName << ": Error sending trigger: error code "
                  << rc << endl;
        else
            *mLog << mName << ": Trigger successfully sent" << endl;

        lmsg::error_t rc2 = mAlarmClient.setAlarm(mMCF_ad, mMCF_ah);
        if (rc2)
            *mLog << mName << ": Error setting alarm for MCF: error code "
                  << rc2 << std::endl;
        else
            *mLog << mName << ": Alarm MCF successfully set" << std::endl;
    }

    if (mDebug > 0) {
        *mLog << mName << ": Exited MCFproblemActions" << std::endl;
    }
} // END: MCFproblemActions


//
// D0problemActions(Time now)
//
void ServoMon::D0problemActions(Time now)
{
    // increment counter
    ++mD0counter;

    // set timestamp
    TimeStr(now, mD0timestamp, mTstampFormat.c_str());
    
    // Set up trigger
    trig::TrigRslt tout(mTrigType.c_str(),mTrigSubtypeD0.c_str(),0);
    tout.setIfos(mIFOname.c_str());
    tout.setTime(now);

    if (mDebug > 0) {
        *mLog << mName << ": Debug trigger of type '"
              << mTrigType << "', subtype '"
              << mTrigSubtypeD0 << "'" << endl;
        cerr << mName << ": Debug trigger of type '"
             << mTrigType << "', subtype '"
             << mTrigSubtypeD0 << "'" << endl;

        lmsg::error_t rc2 = mAlarmClient.setAlarm(mD0_ad, mD0_ah);
        std::cerr << mName << ": setAlarm returned " << rc2 << std::endl;
        if (rc2)
            *mLog << mName << ": Error setting alarm for DARM_CTRL violin 340Hz-345Hz: error code "
                  << rc2 << std::endl;
        else
            *mLog << mName << ": Alarm DARM_CTRL violin 340Hz-345Hz successfully set" << std::endl;

    } else {
        // send trigger to metaDB and raise alarm
        tout.setDisposition(trig::d_metaDB + trig::d_alarm);
            
        *mLog << mName << ": Sending trigger of type '"
              << mTrigType << "', subtype '"
              << mTrigSubtypeD0 << "'" << endl;
        
        lmsg::error_type rc = sendTrigger(tout);
        if (rc)
            *mLog << mName << ": Error sending trigger: error code "
                  << rc << endl;
        else
            *mLog << mName << ": Trigger successfully sent" << endl;

        lmsg::error_t rc2 = mAlarmClient.setAlarm(mD0_ad, mD0_ah);
        if (rc2)
            *mLog << mName << ": Error setting alarm for DARM_CTRL violin 340Hz-345Hz: error code "
                  << rc2 << std::endl;
        else
            *mLog << mName << ": Alarm DARM_CTRL violin 340Hz-345Hz successfully set" << std::endl;
    }
} // END: D0problemActions


//
// D0pProblemActions(Time now)
//
void ServoMon::D0pProblemActions(Time now)
{
    // increment counter
    ++mD0pCounter;

    // set timestamp
    TimeStr(now, mD0pTimestamp, mTstampFormat.c_str());
    
    // Set up trigger
    trig::TrigRslt tout(mTrigType.c_str(),mTrigSubtypeD0plus.c_str(),0);
    tout.setIfos(mIFOname.c_str());
    tout.setTime(now);

    if (mDebug > 0) {
        *mLog << mName << ": Debug trigger of type '"
              << mTrigType << "', subtype '"
              << mTrigSubtypeD0plus << "'" << endl;
        cerr << mName << ": Debug trigger of type '"
             << mTrigType << "', subtype '"
             << mTrigSubtypeD0plus << "'" << endl;
    } else {
        // send trigger to metaDB and raise alarm
        tout.setDisposition(trig::d_metaDB + trig::d_alarm);
            
        *mLog << mName << ": Sending trigger of type '"
              << mTrigType << "', subtype '"
              << mTrigSubtypeD0plus << "'" << endl;
        
        lmsg::error_type rc = sendTrigger(tout);
        if (rc)
            *mLog << mName << ": Error sending trigger: error code "
                  << rc << endl;
        else
            *mLog << mName << ": Trigger successfully sent" << endl;


        lmsg::error_t rc2 = mAlarmClient.setAlarm(mD0p_ad, mD0p_ah);
        if (rc2)
            *mLog << mName << ": Error setting alarm for DARM_CTRL 345Hz-350Hz: error code "
                  << rc2 << std::endl;
        else
            *mLog << mName << ": Alarm DARM_CTRL 345Hz-350Hz successfully set" << std::endl;
    }
} // END: D0pProblemActions


//
// D1problemActions(Time now)
//
void ServoMon::D1problemActions(Time now)
{
    // increment counter
    ++mD1counter;

    // set timestamp
    TimeStr(now, mD1timestamp, mTstampFormat.c_str());
    
    // Set up trigger
    trig::TrigRslt tout(mTrigType.c_str(),mTrigSubtypeD1.c_str(),0);
    tout.setIfos(mIFOname.c_str());
    tout.setTime(now);

    if (mDebug > 0) {
        *mLog << mName << ": Debug trigger of type '"
              << mTrigType << "', subtype '"
              << mTrigSubtypeD1 << "'" << endl;
        cerr << mName << ": Debug trigger of type '"
             << mTrigType << "', subtype '"
             << mTrigSubtypeD1 << "'" << endl;
    } else {
        // send trigger to metaDB and raise alarm
        tout.setDisposition(trig::d_metaDB + trig::d_alarm);
            
        *mLog << mName << ": Sending trigger of type '"
              << mTrigType << "', subtype '"
              << mTrigSubtypeD1 << "'" << endl;
        
        lmsg::error_type rc = sendTrigger(tout);
        if (rc)
            *mLog << mName << ": Error sending trigger: error code "
                  << rc << endl;
        else
            *mLog << mName << ": Trigger successfully sent" << endl;

        lmsg::error_t rc2 = mAlarmClient.setAlarm(mD1_ad, mD1_ah);
        if (rc2)
            *mLog << mName << ": Error setting alarm for DARM_CTRL violin 680Hz-690Hz: error code "
                  << rc2 << std::endl;
        else
            *mLog << mName << ": Alarm DARM_CTRL violin 680Hz-690Hz successfully set"
                  << std::endl;
    }
} // END: D1problemActions


//
// D2problemActions(Time now)
//
void ServoMon::D2problemActions(Time now)
{
    // increment counter
    ++mD2counter;

    // set timestamp
    TimeStr(now, mD2timestamp, mTstampFormat.c_str());
    
    // Set up trigger
    trig::TrigRslt tout(mTrigType.c_str(),mTrigSubtypeD1.c_str(),0);
    tout.setIfos(mIFOname.c_str());
    tout.setTime(now);

    if (mDebug > 0) {
        *mLog << mName << ": Debug trigger of type '"
              << mTrigType << "', subtype '"
              << mTrigSubtypeD2 << "'" << endl;
        cerr << mName << ": Debug trigger of type '"
             << mTrigType << "', subtype '"
             << mTrigSubtypeD2 << "'" << endl;
    } else {
        // send trigger to metaDB and raise alarm
        tout.setDisposition(trig::d_metaDB + trig::d_alarm);
            
        *mLog << mName << ": Sending trigger of type '"
              << mTrigType << "', subtype '"
              << mTrigSubtypeD2 << "'" << endl;
        
        lmsg::error_type rc = sendTrigger(tout);
        if (rc)
            *mLog << mName << ": Error sending trigger: error code "
                  << rc << endl;
        else
            *mLog << mName << ": Trigger successfully sent" << endl;

        lmsg::error_t rc2 = mAlarmClient.setAlarm(mD2_ad, mD2_ah);
        if (rc2)
            *mLog << mName << ": Error setting alarm for DARM_CTRL 400Hz-450Hz: error code "
                  << rc2 << std::endl;
        else
            *mLog << mName << ": Alarm DARM_CTRL 400Hz-450Hz successfully set"
                  << std::endl;
    }
} // END: D2problemActions


//
// statisticsActions()
//
void ServoMon::statisticsActions(const Time & now)
{
    mMCFdata.push_back(mOsclist[mIFOname + ":MC_F_peaking"]->watchedQuantity());
    mD0data.push_back(mOsclist[mIFOname + ":DARM_CTRL_violin_0"]->watchedQuantity());
    mD0plusdata.push_back(mOsclist[mIFOname + ":DARM_CTRL_violin_0_plus"]->watchedQuantity());
    mD1data.push_back(mOsclist[mIFOname + ":DARM_CTRL_violin_1"]->watchedQuantity());
    mD2data.push_back(mOsclist[mIFOname + ":DARM_CTRL_400_450"]->watchedQuantity());

    if (now.getS() % 5 == 0) {
        fprintf(stderr, "ServoMon statistics for power:\n");
        fprintf(stderr, "\tMCF: Mean = %12.8e;\tStdDev = %12.8e\n",
                mean(mMCFdata), stdDev(mMCFdata));
        fprintf(stderr, "\tD0:  Mean = %12.8e;\tStdDev = %12.8e\n",
                mean(mD0data), stdDev(mD0data));
        fprintf(stderr, "\tD0p: Mean = %12.8e;\tStdDev = %12.8e\n",
                mean(mD0plusdata), stdDev(mD0plusdata));
        fprintf(stderr, "\tD1:  Mean = %12.8e;\tStdDev = %12.8e\n",
                mean(mD1data), stdDev(mD1data));
        fprintf(stderr, "\tD2:  Mean = %12.8e;\tStdDev = %12.8e\n",
                mean(mD1data), stdDev(mD1data));
    }
} // END: statisticsActions()


//
// printCounters()
//
void ServoMon::printCounters(void) const
{
    *mLog << "Counters and timestamps:" << std::endl;
    *mLog << "\tMCF: " << mMCFcounter << "; " << mMCFtimestamp << std::endl;
    *mLog << "\tD0:  " << mD0counter << "; " << mD0timestamp << std::endl;
    *mLog << "\tD0+: " << mD0pCounter << "; " << mD0pTimestamp << std::endl;
    *mLog << "\tD1:  " << mD1counter << "; " << mD1timestamp << std::endl;
    *mLog << "\tD2:  " << mD2counter << "; " << mD2timestamp << std::endl;
}// END: printCounters()


//
// ProcessData()
//
void ServoMon::ProcessData(void)
{
   // check stride alignment -- copied from PSLmon
    // this is attempt at fixing problem with zeros at start of trend frames
    Time tEnd = mIn.getCurrentTime();
    int tSec = tEnd.getS();
    int iStride = int(mIn.getStride());

    ++mFrameCount;
    mTotalRunTime += getDacc().getStride();

    if (mDebug > 0) {
        *mLog << endl << mName << ": Entered ProcessData() with mFrameCount = "
              << mFrameCount << endl;
    }

    Time now = mIn.getFillTime();

    TimeStr(now, mTimestamp, mTstampFormat.c_str());

    *mLog << endl << mName << ": Time = " << mTimestamp
          << "; GPS " << now << endl;

    // Stuff to do the first time around
    if (mFrameCount == 1) {
        initTSWindows(now);
    }

    bool MCFproblem_p    = false;
    bool D0problem_p     = false;
    bool D0plusProblem_p = false;
    bool D1problem_p     = false;
    bool D2problem_p     = false;

    // informational output only every 15 strides
    if (mDebug > 0 || mFrameCount % 15 == 0) {
        *mLog << "Processing event number " << mFrameCount << endl;

        if (mDebug > 2)
            cerr << mName << ": event number " << mFrameCount << endl;
    }

    if (mFrameCount == 1) { 
      *mLog << "Will check MC_F and DARM_CTRL power spectrum every 20 seconds"
            << endl;
      *mLog << "Will issue warning no more often than every 5 minutes"
            << endl;
      *mLog << "if excess power present in MC_F (3-7kHz),"
            << " DARM_CTRL (340-345Hz, 345-355Hz, 680-690Hz) continuously"
            << endl << endl;
      *mLog << "Starting " << mIFOname << " monitor with ";
    }

    //
    // Check for problems, and send triggers if necessary
    //

    // MC_problem
    if (mOsclist.satisfied(mIFOname + ":MC_problem")) {
        if (mDebug > 0 || mFrameCount == 1) {
            *mLog << " problem in MC at time " << now.getS() << endl;
            cerr << mName << ": problem in MC: at time " << now.getS() << endl;
        }

        mMCFProblemPWindow.append(1, now);

        mMCFTriggers.append(1, now);

        // Make sure we don't send trigger more often than once every
        // 5 minutes
        if (mFrameCount - mMCFLastProb > 15 || mFrameCount == 1) {
            *mLog << mIFOname
                  << ": MC_F showing power excess at high freqs. at time "
                  << now.getS() << endl;
            mMCFLastProb = mFrameCount;
            MCFproblem_p = true;
        }
    } else {
        if (mDebug > 0 || mFrameCount == 1)
            *mLog << "no problem in MC" << endl;
        
        mMCFProblemPWindow.append(0,now);
        mMCFTriggers.append(0,now);
    }

    // DARM_problem_0
    if (mOsclist.satisfied(mIFOname + ":DARM_problem_0")) {
        if (mDebug > 0 || mFrameCount == 1) {
            *mLog << mName << ": problem 0 in DARM_CTRL: at time "
                  << now.getS() << endl;
            cerr << mName << ": problem 0 in DARM_CTRL: at time "
                 << now.getS() << endl;
        }

        mD0Triggers.append(1,now);

        // Make sure we don't send trigger more often than once every
        // 5 minutes
        if (mFrameCount - mD0LastProb > 15 || mFrameCount == 1) {
            *mLog << mIFOname
                  << ": DARM_CTRL showing excess power in violin "
                  << "fundamental mode (340-345Hz) at time "
                  << now.getS() << endl;
            mD0LastProb = mFrameCount;
            D0problem_p = true;
        }        
    } else {
        mD0Triggers.append(0,now);
    }

    // DARM_problem_0_plus
    if (mOsclist.satisfied(mIFOname + ":DARM_problem_0_plus")) {
        if (mDebug > 0 || mFrameCount == 1) {
            *mLog << mName << ": problem 0plus in DARM_CTRL: at time "
                  << now.getS() << endl;
            cerr << mName << ": problem 0plus in DARM_CTRL: at time "
                 << now.getS() << endl;
        }

        mD0pTriggers.append(1,now);

        // Make sure we don't send trigger more often than once every
        // 5 minutes
        if (mFrameCount - mD0pLastProb > 15 || mFrameCount == 1) {
            *mLog << mIFOname
                  << ": DARM_CTRL showing excess power in "
                  << "345Hz-355Hz band at time "
                  << now.getS() << endl;
            mD0pLastProb = mFrameCount;
            D0plusProblem_p = true;
        }        
    } else {
        mD0pTriggers.append(0,now);
    }

    // DARM_problem_1
    if (mOsclist.satisfied(mIFOname + ":DARM_problem_1")) {
        if (mDebug > 0 || mFrameCount == 1) {
            *mLog << mName << ": problem 1 in DARM_CTRL: at time "
                  << now.getS() << endl;
            cerr << mName << ": problem 1 in DARM_CTRL: at time "
                 << now.getS() << endl;
        }

        mD1Triggers.append(1,now);
        
        // Make sure we don't send trigger more often than once every
        // 5 minutes
        if (mFrameCount - mD1LastProb > 15 || mFrameCount == 1) {
            *mLog << " DARM_CTRL showing power excess at "
                  << "680-690Hz at time "
                  << now.getS() << endl;
            mD1LastProb = mFrameCount;
            D1problem_p = true;
        }        
    } else {
        mD1Triggers.append(0,now);
    }

    // DARM_problem_2 400-450Hz
    if (mOsclist.satisfied(mIFOname + ":DARM_problem_2")) {
        if (mDebug > 0 || mFrameCount == 1) {
            *mLog << mName << ": problem 2 in DARM_CTRL: at time "
                  << now.getS() << endl;
            cerr << mName << ": problem 2 in DARM_CTRL: at time "
                 << now.getS() << endl;
        }

        mD2Triggers.append(1,now);
        
        // Make sure we don't send trigger more often than once every
        // 5 minutes
        if (mFrameCount - mD1LastProb > 15 || mFrameCount == 1) {
            *mLog << " DARM_CTRL showing power excess in "
                  << "400-450Hz band at time "
                  << now.getS() << endl;
            mD1LastProb = mFrameCount;
            D1problem_p = true;
        }        
    } else {
        mD2Triggers.append(0,now);
    }

    // Add current power in violin mode band to PowerWindows
    updatePowerWindows(now);

    // Write out violin mode power to log file
    appendToLog(now);
    
    if (mDebug > 2) {
        OperStateCondList::iterator iter = mOsclist.begin(); 
        for (; iter != mOsclist.end(); ++iter) {
            cout << "'" << (*iter).first << "': watchedQuantity() = ";
            printf("%12.8e\n", (*iter).second->watchedQuantity());
            if ((*iter).second->satisfied() == true)
                cout << "satisfied" << endl;
            else
                cout << "NOT satisfied" << endl;
        }
        cout << endl;
    }

    //
    // If problem arises send trigger.
    //
    if (MCFproblem_p == true) 
        MCFproblemActions(now);

    if (D0problem_p == true)
        D0problemActions(now);

    if (D0plusProblem_p == true)
        D0pProblemActions(now);

    if (D1problem_p == true)
        D1problemActions(now);

    if (D2problem_p == true)
        D2problemActions(now);

    if (mDebug > 0) {
        std::cerr << mName << ": FOOBAR -- mDispfileName = "
                  << mDispfileName << std::endl;
    }

    // Write out HTML page every 300 = 15*20 seconds
    if (mFrameCount % 15 == 0 || mFrameCount == 1) {

        // Open display file with overwrite mode
        std::ofstream dispfile(mDispfileName.c_str(), ios::out);

        if (!dispfile) {
            std::cerr << mName << ": could not open dispfile "
                      << mDispfileName << std::endl;
            std::cerr << "          reverting to stdout" << std::endl;
            mDisplay = &std::cout;
        } else {
            mDisplay = &dispfile;
        }

        writeStatusPage();
        dispfile.close();
    }

    // Fill in TSeries for DMT viewer
    mMCFProblemPWindow.fillTSeries(mMCFProblemTS);
    mPowerWindowMCF.fillTSeries(mPowerTSMCF);
    mPowerWindowD0.fillTSeries(mPowerTSD0);
    mPowerWindowD0plus.fillTSeries(mPowerTSD0plus);
    mPowerWindowD1.fillTSeries(mPowerTSD1);
    mPowerWindowD2.fillTSeries(mPowerTSD2);

    // Update trend data
    mPowerWindowMCF.fillVector(mTmpVec, 3);
    if (mDebug > 0)
        cout << "mPowerWindowMCF.length() = " << mPowerWindowMCF.length() << endl;
    mTrend.trendData(mTrendChanNames["MCF"].c_str(), now,
                     mean(mTmpVec));
    mPowerWindowD0.fillVector(mTmpVec, 3);
    mTrend.trendData(mTrendChanNames["D0"].c_str(), now,
                     mean(mTmpVec));
    mPowerWindowD0plus.fillVector(mTmpVec, 3);
    mTrend.trendData(mTrendChanNames["D0p"].c_str(), now,
                     mean(mTmpVec));
    mPowerWindowD1.fillVector(mTmpVec, 3);
    mTrend.trendData(mTrendChanNames["D1"].c_str(), now,
                     mean(mTmpVec));
    mPowerWindowD2.fillVector(mTmpVec, 3);
    mTrend.trendData(mTrendChanNames["D2"].c_str(), now,
                     mean(mTmpVec));

    if (mDebug > 0)
        cout << "  updating trend data" << endl;

    if (mDebug > 0) {
        cout << "now.getS() = " << now.getS() << endl;
        cout << "now.getS() % 60 = " << now.getS() % 60 << endl;
        cout << "now.getS() % 3600 = " << now.getS() % 3600 << endl;
    }

    // Update trends
    mTrend.Update(now);

    // checkpoint strip charts every 10 minutes
    if (mFrameCount > 10 && (now.getS() % 600) ==0)
        saveAllTSWindows();

    if (mDebug > 0) {
        statisticsActions(now);

        // print out counters and timestamps
        printCounters();
    }

    return;
} // END: ProcessData()


//
// HTML display page stuff
//
void ServoMon::htmlHead(void) const
{
    *mDisplay << "<!doctype html public \"-//W3C//DTD HTML 4.0 Transitional//EN\""
              << endl;
    *mDisplay << "\"http://www.w3.org/TR/REC-html40/loose.dtd\">"
              << endl;

    *mDisplay << "<html><head>" << endl;
    *mDisplay << "<!-- dwchin@umich.edu -->" << endl;
    *mDisplay << "<meta http-equiv=\"Refresh\" content=\"" << mHTMLupdateInterval
              << "\">" << endl;
    *mDisplay << "<style type=\"text/css\">" << endl;
    *mDisplay << "   <!-- body { color: black; background: white; " << endl;
    *mDisplay << "               font-family: Helvetica, Arial, sans-serif; " << endl
              << "               font-weight: normal; } " << endl
              << "        h1   { font-weight: bold; }" << endl
              << "        h2   { font-weight: bold; }" << endl
              << "        h3   { font-weight: bold; }" << endl
              << "        h4   { font-weight: bold; }" << endl
              << "        h5   { font-weight: bold; }" << endl
              << "    -->" << endl;
    *mDisplay << "</style>" << endl;
    *mDisplay << "<title>ServoMon: Current Status for " << mIFOname;
    if (mIFOname == "H1") {
        *mDisplay << " (Hanford 4k)";
    } else if (mIFOname == "H2") {
        *mDisplay << " (Hanford 2k)";
    } else if (mIFOname == "L1") {
        *mDisplay << " (Livingston 4k)";
    } else {
        // should never reach here since mIFOname is checked before
        std::string eObj = "Unknown IFO: ";
        eObj += mIFOname;
        cerr << eObj << endl;
        throw runtime_error(eObj);
    }
    *mDisplay << "</title>" << endl;

    *mDisplay << "</head>" << endl;
    
    *mDisplay << "<body><h1>ServoMon: MC_F High Frequency "
              << "and DARM_CTRL Violin Mode Noise Status for " << mIFOname;
    if (mIFOname == "H1") {
        *mDisplay << " (Hanford 4k)";
    } else if (mIFOname == "H2") {
        *mDisplay << " (Hanford 2k)";
    } else if (mIFOname == "L1") {
        *mDisplay << " (Livingston 4k)";
    } else {
        // should never reach here since mIFOname is checked before
        std::string eObj = "Unknown IFO: ";
        eObj += mIFOname;
        cerr << eObj << endl;
        throw runtime_error(eObj);
    }
    *mDisplay << "</h1>" << endl;
} // END: htmlHead()

void ServoMon::htmlFoot(void) const
{
    *mDisplay << "<p>&nbsp;</p><hr>" << endl;
    *mDisplay << "Contact: <a href=\"mailto:dwchin@umich.edu\">David Chin <tt>&lt;dwchin@umich.edu&gt;</tt></a> 734-730-1274"
              << endl;
    *mDisplay << "</body></html>" << endl;
}

void ServoMon::writeStatusPage(void)
{
    if (mDebug > 0) {
        std::cerr << mName << ": writing status page..." << std::endl;
    }
    
    htmlHead();

    *mDisplay << "<h3>Last update: " << mTimestamp << "</h3>" << std::endl;

    *mDisplay << "<p>ServoMon (" << mIFOname << ") has been running "
              << "for " << mTotalRunTime.GetSecs() << " seconds ("
              << (double)(mTotalRunTime.GetSecs())/3600.
              << " hrs)</p>" << std::endl;

    *mDisplay << "<p><font size=\"-1\"><em>Updates every 5 mins ";
    *mDisplay << "($Revision: 1.25 $)</em></font></p>" << std::endl;

    // comment
    *mDisplay << "<p><i>Shows raw trigger rates for ServoMon band-limited power "
              << "conditions.  While GDS MetaDB triggers are throttled down to "
              << "1 per 5 minutes, the trigger rates here are not throttled.  "
              << "<b>However</b>, these rates can only have a maximum of 0.05 Hz "
              << "because the data is read in 20 second strides.</i></p>" << std::endl;
    *mDisplay << "<p>Negative values in the tables below indicate that data is "
              << "unavailable.</p>" << std::endl;
    *mDisplay << "<p><font color=\"red\">NOTE:</font> cumulative rates are "
              << "only since this instance of the monitor began running.</p>" << std::endl;

    // compute trigger rates since monitor started running
    double trigRateMCF, trigRateD0, trigRateD0p, trigRateD1, trigRateD2;

    trigRateMCF = static_cast<double>(mMCFcounter) / (static_cast<double>(mFrameCount) * 20.);
    trigRateD0  = static_cast<double>(mD0counter) / (static_cast<double>(mFrameCount) * 20.);
    trigRateD0p = static_cast<double>(mD0pCounter) / (static_cast<double>(mFrameCount) * 20.);
    trigRateD1  = static_cast<double>(mD1counter) / (static_cast<double>(mFrameCount) * 20.);
    trigRateD2  = static_cast<double>(mD2counter) / (static_cast<double>(mFrameCount) * 20.);

    // print out trig rate tables here.

    *mDisplay << "<table width=\"85%\" align=\"center\" summary=\"ServoMon trigger rates in Hz\">";
    *mDisplay << "<caption>Trigger rates in Hz</caption>";
    *mDisplay << "<tr colspan=\"2\"><td align=\"center\">";
    makeTrigRateTable("MC_F 3kHz-7kHz noise", trigRateMCF,
                      triggerRate(mMCFTriggers, 2),
                      triggerRate(mMCFTriggers, 6),
                      triggerRate(mMCFTriggers, 12),
                      triggerRate(mMCFTriggers, 24));
    *mDisplay << "</td></tr><tr><td align=\"center\">";

    makeTrigRateTable("DARM_CTRL 340Hz-345Hz noise", trigRateD0,
                      triggerRate(mD0Triggers, 2),
                      triggerRate(mD0Triggers, 6),
                      triggerRate(mD0Triggers, 12),
                      triggerRate(mD0Triggers, 24));

    *mDisplay << "</td><td align=\"center\">";
    makeTrigRateTable("DARM_CTRL 345Hz-355Hz noise", trigRateD0p,
                      triggerRate(mD0pTriggers, 2),
                      triggerRate(mD0pTriggers, 6),
                      triggerRate(mD0pTriggers, 12),
                      triggerRate(mD0pTriggers, 24));
    
    *mDisplay << "</td></tr><tr><td align=\"center\">";
    makeTrigRateTable("DARM_CTRL 680Hz-690Hz noise", trigRateD1,
                      triggerRate(mD1Triggers, 2),
                      triggerRate(mD1Triggers, 6),
                      triggerRate(mD1Triggers, 12),
                      triggerRate(mD1Triggers, 24));

    *mDisplay << "</td><td align=\"center\">";
    makeTrigRateTable("DARM_CTRL 400Hz-450Hz noise", trigRateD2,
                      triggerRate(mD2Triggers, 2),
                      triggerRate(mD2Triggers, 6),
                      triggerRate(mD2Triggers, 12),
                      triggerRate(mD2Triggers, 24));
    *mDisplay << "</td></tr></table>";

    htmlFoot();

    if (mDebug > 0) {
        std::cerr << mName << ": done writing status page." << std::endl;
    }
} // END: writeStatusPage()

//
// ServoMon::writeIndexPage()
//
void ServoMon::writeIndexPage(void)
{
    mIndexPage.open(mIndexPageName.c_str(), ios::out);

    if (!mIndexPage) {
         char tmpstr[256];
        sprintf(tmpstr,
                "%s: Error opening index.html file for writing. Aborting.",
                mName.c_str());
        std::string eObj(tmpstr);
        std::cerr << eObj << std::endl;
        throw runtime_error(eObj);
    } else {
         mIndexPage << "<!doctype html public \"-//w3c//dtd html 4.0 transitional//en\">"
                   << std::endl;
        mIndexPage << "<html><head> " << std::endl;
        mIndexPage << "<!-- dwchin@umich.edu -->" << std::endl;
        mIndexPage << "<meta http-equiv=\"Content-Type\" content=\"text/html; charset=iso-8859-1\">";
        mIndexPage << "<style type=\"text/css\">" << std::endl;
        mIndexPage << "   <!-- body { color: black; background: white; "
                   << "               font-family: Helvetica, Arial, sans-serif; "
                   << "               font-weight: normal; } "
                   << "        h1   { font-weight: bold; }"
                   << "        h2   { font-weight: bold; }"
                   << "        h3   { font-weight: bold; }"
                   << "        h4   { font-weight: bold; }"
                   << "        h5   { font-weight: bold; }"
                   << " -->" << std::endl
                   << "</style>" << std::endl;
        mIndexPage << "<title>" << mName << "</title>" << std::endl;
        mIndexPage << "</head><body><center>" << std::endl;
        mIndexPage << "<h2>Global summary for " << mName
                   << " DMT Monitor running at ";

        if (mIFOname == "H1" || mIFOname == "H2")
            mIndexPage << "Hanford";
        else if (mIFOname == "L1")
            mIndexPage << "Livingston";
        else
            mIndexPage << "Timbuctoo";
             
        mIndexPage << "</h2></center>" << std::endl;

        mIndexPage << "<h3>" << mIFOname << " Interferometer (";

        if (mIFOname == "H1" || mIFOname == "L1")
            mIndexPage << "4";
        else if (mIFOname == "H2")
            mIndexPage << "2";

        mIndexPage << "km)</h3>" << std::endl;

        mIndexPage << "<ul>";
        mIndexPage << "<li><a href=\"ServoMon_Status.html\">";
        mIndexPage << "ServoMon status page</a></li>";
        mIndexPage << "<li><a href=\"ServoMon_Log.txt\">";
        mIndexPage << "Raw log file</a></li></ul>";

        mIndexPage << "(Time series of band-limited RMS/power "
                   << "may be seen via the GDS&nbsp;DMT Viewer)";
        mIndexPage << "<p>&nbsp;</p><hr><p>Contact: "
                   << "<a href=\"mailto:dwchin@umich.edu\">David Chin "
                   << "&lt;dwchin@umich.edu&gt;</a>";
        mIndexPage << "</body></html>" << std::endl;
    }

    mIndexPage.close();
} // END: writeIndexPage()

//
// setTSWindowStuff()
//
void ServoMon::setTSWindowStuff(void)
{
    //
    // Set window params
    //

    // strings for savefile names
    std::string chartstr = mHtmlDir + mIFOname + "_StripChart";
    const Interval dt((long)mStride,0);

    //mMCFProblemPWindow.setDebug(mDebug);
    mMCFProblemPWindow.setDt(dt);
    mMCFProblemPWindow.setName(mIFOname + mProblemPstr);
    mMCFProblemPWindow.setSaveFileName(chartstr + "_MCF_Problem.dat");

    //mPowerWindowMCF.setDebug(mDebug);
    mPowerWindowMCF.setDt(dt);
    mPowerWindowMCF.setName(mIFOname + mMCFstr);
    mPowerWindowMCF.setSaveFileName(chartstr + "_MCF_Power.dat");

    //mPowerWindowD0.setDebug(mDebug);
    mPowerWindowD0.setDt(dt);
    mPowerWindowD0.setName(mIFOname + mD0str);
    mPowerWindowD0.setSaveFileName(chartstr + "_340-345Hz_Power.dat");

    //mPowerWindowD0plus.setDebug(mDebug);
    mPowerWindowD0plus.setDt(dt);
    mPowerWindowD0plus.setName(mIFOname + mD0plusstr);
    mPowerWindowD0plus.setSaveFileName(chartstr + "_345-355Hz_Power.dat");

    //mPowerWindowD1.setDebug(mDebug);
    mPowerWindowD1.setDt(dt);
    mPowerWindowD1.setName(mIFOname + mD1str);
    mPowerWindowD1.setSaveFileName(chartstr + "_680-690Hz_Power.dat");

    mPowerWindowD2.setDt(dt);
    mPowerWindowD2.setName(mIFOname + mD2str);
    mPowerWindowD2.setSaveFileName(chartstr + "_400-450Hz_Power.dat");


    // The trigger rate TSWindows
    
    std::string trigstr = mHtmlDir + mIFOname + "_Trigs_";
    mMCFTriggers.setDt(dt);
    mMCFTriggers.setName(mIFOname + "MCF_Problem.dat");
    mMCFTriggers.setSaveFileName(trigstr + "MCF_Problem.dat");
    
    mD0Triggers.setDt(dt);
    mD0Triggers.setName(mIFOname + "DARM_CTRL-340-345Hz_Power.dat");
    mD0Triggers.setSaveFileName(trigstr + "DARM_CTRL-340-345Hz_Power.dat");

    mD0pTriggers.setDt(dt);
    mD0pTriggers.setName(mIFOname + "DARM_CTRL-345-355Hz_Power.dat");
    mD0pTriggers.setSaveFileName(trigstr + "DARM_CTRL-345-355Hz_Power.dat");

    mD1Triggers.setDt(dt);
    mD1Triggers.setName(mIFOname + "DARM_CTRL-680-690Hz_Power.dat");
    mD1Triggers.setSaveFileName(trigstr + "DARM_CTRL-680-690Hz_Power.dat");

    mD2Triggers.setDt(dt);
    mD2Triggers.setName(mIFOname + "DARM_CTRL-400-450Hz_Power.dat");
    mD2Triggers.setSaveFileName(trigstr + "DARM_CTRL-400-450Hz_Power.dat");
} // END: setTSWindowStuff()


//
// initTSWindows()
//
void ServoMon::initTSWindows(const Time & now)
{
    if (mDebug > 0) {
        std::cout << "ServoMon::initTSWindows(): now = " << now
                  << std::endl;
    }

    // 12 hrs for strip charts
    mMCFProblemPWindow.fillMaybe(now - Interval(43200,0));
    mPowerWindowMCF.fillMaybe(now - Interval(43200,0));
    mPowerWindowD0.fillMaybe(now - Interval(43200,0));
    mPowerWindowD0plus.fillMaybe(now - Interval(43200,0));
    mPowerWindowD1.fillMaybe(now - Interval(43200,0));
    mPowerWindowD2.fillMaybe(now - Interval(43200,0));

    // 24 hrs for triggers
    mMCFTriggers.fillMaybe(now - Interval(86400,0));
    mD0Triggers.fillMaybe(now - Interval(86400,0));
    mD0pTriggers.fillMaybe(now - Interval(86400,0));
    mD1Triggers.fillMaybe(now - Interval(86400,0));
    mD2Triggers.fillMaybe(now - Interval(86400,0));

    if (mDebug > 0) {
        std::cout << "ServoMon::initTSWindows(): done"
                  << std::endl;
    }
} // END: initTSWindows()

//
// restoreAllTSWindows()
//
void ServoMon::restoreAllTSWindows(void)
{
    if (mDebug > 0) {
        std::cout << "ServoMon::restoreAllTSWindows(): entered" << std::endl;
    }

    mMCFProblemPWindow.restore();
    mPowerWindowMCF.restore();
    mPowerWindowD0.restore();
    mPowerWindowD0plus.restore();
    mPowerWindowD1.restore();
    mPowerWindowD2.restore();

    mMCFTriggers.restore();
    std::cout << "Triggers TSWindows: MCF ";
    mD0Triggers.restore();
    std::cout << "D0 ";
    mD0pTriggers.restore();
    std::cout << "D0+ ";
    mD1Triggers.restore();
    std::cout << "D1 ";
    mD2Triggers.restore();
    std::cout << "D2 ";

    if (mDebug > 0) {
        std::cout << "ServoMon::restoreAllTSWindows(): done" << std::endl;
    }
} // END: restoreAllTSWindows()


//
// saveAllTSWindows()
//
void ServoMon::saveAllTSWindows(void)
{
    if (mDebug > 0) {
        std::cout << "ServoMon::saveAllTSWindows(): entered" << std::endl;
    }

    mMCFProblemPWindow.save();
    mPowerWindowMCF.save();
    mPowerWindowD0.save();
    mPowerWindowD0plus.save();
    mPowerWindowD1.save();
    mPowerWindowD2.save();

    mMCFTriggers.save();
    mD0Triggers.save();
    mD0pTriggers.save();
    mD1Triggers.save();
    mD2Triggers.save();

    if (mDebug > 0) {
        std::cout << "ServoMon::saveAllTSWindows(): done" << std::endl;
    }
} // END: saveAllTSWindows()


void ServoMon::makeTrigRateTable(const std::string & name, 
                                 double cumulRate,
                                 double twoHrRate,
                                 double sixHrRate,
                                 double twelveHrRate,
                                 double oneDayRate)
{
   mDisplay->setf(ios::scientific);
   mDisplay->precision(6);
   *mDisplay << "<table summary=\"" << name << "\" width=\"85%\" align=\"center\" border=\"1\">"
             << "<caption><b>" << name << "</b></caption>"
             << "<tr><td valign=\"middle\">Cumulative trigger rate: </td><td align=\"right\">"  
             << cumulRate << "</td></tr>" << std::endl
             << "<tr><td valign=\"middle\">2 Hr trigger rate: </td><td align=\"right\">"
             << twoHrRate << "</td></tr>" << std::endl
             << "<tr><td valign=\"middle\">6 Hr trigger rate: </td><td align=\"right\">"
             << sixHrRate << "</td></tr>" << std::endl
             << "<tr><td valign=\"middle\">12 Hr trigger rate: </td><td align=\"right\">"
             << twelveHrRate << "</td></tr>" << std::endl
             << "<tr><td valign=\"middle\">24 Hr trigger rate: </td><td align=\"right\">"
             << oneDayRate << "</td></tr>" << std::endl
             << "</table>" << std::endl
             << "<p>&nbsp;</p>" << std::endl;
} // END: makeTrigRateTable()


//
// Return trigger rate in Hz
//
double ServoMon::triggerRate(const TSWindow<int> & triggers, size_t nhours, size_t nminutes, size_t nsecs)
{
    if (mDebug > 0) {
        std::cout << "ServoMon::triggerRate(): entered" << std::endl;
    }
    
    std::vector<int> trigdata;

    size_t total_secs = nhours*3600
                       + nminutes*60
                       + nsecs;

    triggers.fillVector(trigdata, total_secs / 20);

    // if insufficient data, the some element will contain
    // the dummy value: return a negative number

    // gah! do a linear search for dummy values
    // do it from the front on the assumption that dummy
    // values are more likely to occur in past times
    for (size_t i = 0; i < trigdata.size(); ++i)
        if (trigdata[i] == triggers.dummyVal())
            return (triggers.dummyVal());

    // count up number of triggers
    int trigcount(0);
    for (size_t i = 0; i < trigdata.size(); ++i) {
        if (trigdata[i] > 0)
            ++trigcount;
    }

    return (static_cast<double>(trigcount) / static_cast<double>(total_secs));
} // END: triggerRate()

