#include "StrainWatch.hh"
#include "TSeries.hh"
#include "Dacc.hh"
#include "Time.hh"
#include "Interval.hh"
#include <math.h>

#define REMEMBER_FACTORS_SECS 10*60
#define EXP_FACTOR .5

EXECDAT( StrainWatch );

StrainWatch::StrainWatch( int argc, const char* argv[] )
	: DatEnv( argc, argv ), mOSC( getDacc() )
{
	setDefaultParams();
	readCommandLine( argc, argv );
	updateParams();
	
	strGet = new StrainGet( &strGetParams );
		
	
	
	mOSC.setStride( timeStride );
	mOSC.readConfig( oscFile );
	
	window = new Hanning();
	psd = new PSD( window, 1 );
//	psd1min = new PSD( window, 15 );
	
	getDacc().setStride( timeStride );
	setStrideAlignment( timeStride, 0.0 );
	getDacc().addChannel( asqChannel.c_str(), downsample );
	getDacc().addChannel( darmChannel.c_str(), downsample );
	getDacc().addChannel( darmErrChannel.c_str(), downsample );
	getDacc().addChannel( excChannel.c_str(), downsample );
	
	string serverName = "StrainWatch_" + ifo;
	setServerName( serverName.c_str() );
	
	int datasize = (wings + 1) * (1 << 14) / downsample;
	asqData = new double[ datasize ];
	darmData = new double[ datasize ];
	darmErrData = new double[ datasize ];
	excData = new double[ datasize ];
	cout << "array size: " << datasize << endl;
	
	int factorsDataSize = (int)(REMEMBER_FACTORS_SECS/factorsTime +0.5);
	alphaRe = new double[ factorsDataSize ];
	alphaIm = new double[ factorsDataSize ];
	betaRe = new double[ factorsDataSize ];
	betaIm = new double[ factorsDataSize ];
	alphabetaRe = new double[ factorsDataSize ];
	alphabetaIm = new double[ factorsDataSize ];
	
	
/*
	arave = new double[ factorsDataSize ];
	aiave = new double[ factorsDataSize ];
	brave = new double[ factorsDataSize ];
	biave = new double[ factorsDataSize ];
	abrave = new double[ factorsDataSize ];
	abiave = new double[ factorsDataSize ];	
*/	
	for( int i = 0; i < factorsDataSize; i++ ) {
		alphaRe[i] = alphaIm[i] = betaRe[i] = betaIm[i]
			= alphabetaRe[i] = alphabetaIm[i]
			/*= arave[i] = aiave[i] = brave[i] = biave[i]
			= abrave[i] = abiave[i]*/ = -1;
	}
	
	aveDataSet = false;
}

void StrainWatch::setDefaultParams() {
	verbosity = 100; //!!!change upon production
	configFile = "None";
	timeStride = 1;
	downsample = 1;
	ifo = "";
	debugMode = false;
	lockConditions = "";
	oscFile = "LockLoss.conf";
	
	strGetParams.f = 1144.3;
	strGetParams.To = 0.0625;
	strGetParams.G0Re = -1.162386784321234e-01;
	strGetParams.G0Im = 2.042814618197689e-01;
	strGetParams.D0Re = -4.752798528445400e-01;
	strGetParams.D0Im = 3.397245549164432e-01;
	strGetParams.W0Re = 9.984499156753539e+01;
	strGetParams.W0Im = 5.562482434293072;
	strGetParams.wings = 2;
	wings = 4;
	
	asqChannel = "LSC-AS_Q";
	excChannel = "LSC-DARM_CTRL_EXC_DAQ";
	darmChannel = "LSC-DARM_CTRL";
	darmErrChannel = "LSC-DARM_ERR";

	intervals = 16;

	isFirstRun = true;
}

void StrainWatch::readCommandLine( int argc, const char *argv[] ) {
	if( argc <= 1 ) return; //if there were no arguments, return.

	//put all the arguments into a string
	string arguments = argv[1];
	for( int i = 2; i < argc; i++ ) {
		string add = argv[i];
		arguments += " " + add;
	}

	//send the string to the parameter loader through a stream.
	stringstream ss(arguments);
	string error = loadParams(&ss);
	if( error != "" ) {
		//Error reading parameters from command line. Terminate.
		cerr << "StrainWatch usage error on command line arguments. "
			 << "Error occured for argument " << error << endl;
		dumpUsage();
		exit( 1 );
	}
}

void StrainWatch::readConfigFile() {
	//open the file for reading
	ifstream fs( configFile.c_str(), ios::in );
	if( !fs.is_open() ) {
		//could not open file. Either the file was not there, we did not
		//have read permissions, or some error occured.
		cerr << "StrainWatch fatal error: could not open config file."
			 << endl << "    Tried to use file " << configFile << endl;
	}
	//try to load the parameters from the file.
	string error = loadParams(&fs);
	fs.close();
	if( error != "" ) {
		//Error reading config file. Terminate.
		cerr << "StrainWatch error in config file. "
			 << "Error occured for argument " << error << endl;
		dumpUsage();
		exit( 1 );
	}
}

string StrainWatch::loadParams( istream *s ) {
	string argname, val;

	while( *s >> argname ) {
		if( argname == "-verbosity" ) {
			if( !(*s >> val) ) return argname;
			verbosity = atoi( val.c_str() );
		} else if( argname == "-config" ) {
			if( !(*s >> val) ) return argname;
			configFile = val;
			readConfigFile();
		} else if( argname == "-help" || argname == "-h" ) {
			dumpUsage();
			exit( 1 );
		} else if( argname == "-ifo" ) {
			if( !(*s >> val) ) return argname;
			ifo = val;
		} else if( argname == "-stride" ) {
			if( !(*s >> val) ) return argname;
			timeStride = atoi( val.c_str() );
		} else if( argname.at(0) == '#' ) {
			char comment[256];
			s->getline( comment, 256 );
		} else if( argname == "-debug" ) {
			debugMode = true;
		} else if( argname == "-asq" || argname == "-A" ) {
			if( !(*s >> val) ) return argname;
			asqChannel = val;
		} else if( argname == "-exc" || argname == "-E" ) {
			if( !(*s >> val) ) return argname;
			excChannel = val;
		} else if( argname == "-darm" || argname == "-D" ) {
			if( !(*s >> val) ) return argname;
			darmChannel = val;
		} else if( argname == "-darmerr" || argname == "-R" ) {
			if( !(*s >> val) ) return argname;
			darmErrChannel = val;
		} else if( argname == "-lock" ) {
			if( !(*s >> val) ) return argname;
			lockConditions = val;
		} else if( argname == "-OSCfile" ) {
			if( !(*s >> val) ) return argname;
			oscFile = val;
		} else if( argname == "-downsample" ) {
			if( !(*s >> val) ) return argname;
			downsample = atoi( val.c_str() );
		} else if( !isDatEnvArg( argname.c_str() ) ) {
/*			csdArgs[csdArgCount++] = argname.c_str();		!!!
			if( !(*s >> val) ) return argname;
			csdArgs[csdArgCount++] = val.c_str();
*/		} 
	}
	
	return "";
}

void StrainWatch::updateParams() {
	if( ifo == "" ) {
		if( asqChannel.at(2) == ':' ) {
			ifo = asqChannel.substr(0,2);
		} else {
			cout << "Usage error: You must specify an IFO to monitor" << endl;
			dumpUsage();
			exit( 1 );
		}
	}

	if( asqChannel.at(2) != ':' )
		asqChannel = ifo + ":" + asqChannel;
	if( excChannel.at(2) != ':' )
		excChannel = ifo + ":" + excChannel;
	if( darmChannel.at(2) != ':' )
		darmChannel = ifo + ":" + darmChannel;
	if( darmErrChannel.at(2) != ':' )
		darmErrChannel = ifo + ":" + darmErrChannel;

	if( lockConditions == "" )
		lockConditions = ifo + ":Both_arms_locked_strict_cm";
}

void StrainWatch::dumpUsage() {
	//!!!do this
}

void StrainWatch::ProcessData() {
	Time curTime = getDacc().getFillTime();
	
	cout << "Processing data at " << curTime << endl;

	int factorDataLength = (int)((wings+1)/factorsTime +0.5);
	int allFactorsDataLength = (int)(REMEMBER_FACTORS_SECS/factorsTime +0.5);
	Time historyStartTime = curTime;
	historyStartTime -= REMEMBER_FACTORS_SECS;
	Time minAgo = curTime;
	minAgo -= 60;
//	double fake[0];
	
	if( isFirstRun ) {
		alphaReTS = new TSeries( historyStartTime, factorsTime,
			allFactorsDataLength, alphaRe );
		alphaImTS = new TSeries( historyStartTime, factorsTime,
			allFactorsDataLength, alphaIm );
		betaReTS = new TSeries( historyStartTime, factorsTime,
			allFactorsDataLength, betaRe );
		betaImTS = new TSeries( historyStartTime, factorsTime,
			allFactorsDataLength, betaIm );
		alphabetaReTS = new TSeries( historyStartTime, factorsTime,
			allFactorsDataLength, alphabetaRe );
		alphabetaImTS = new TSeries( historyStartTime, factorsTime,
			allFactorsDataLength, alphabetaIm );

//		hRTS1min = new TSeries( minAgo, factorsTime, 0, fake );

/*	
		araveTS = new TSeries( historyStartTime, factorsTime,
			allFactorsDataLength, arave );
		aiaveTS = new TSeries( historyStartTime, factorsTime,
			allFactorsDataLength, aiave );
		braveTS = new TSeries( historyStartTime, factorsTime,
			allFactorsDataLength, brave );
		biaveTS = new TSeries( historyStartTime, factorsTime,
			allFactorsDataLength, biave );
		abraveTS = new TSeries( historyStartTime, factorsTime,
			allFactorsDataLength, abrave );
		abiaveTS = new TSeries( historyStartTime, factorsTime,
			allFactorsDataLength, abiave );
*/	
		serveData( (ifo + ":DMT-STRW_alpha_re_raw").c_str(), alphaReTS );
		serveData( (ifo + ":DMT-STRW_alpha_im_raw").c_str(), alphaImTS );
		serveData( (ifo + ":DMT-STRW_beta_re_raw").c_str(), betaReTS );
		serveData( (ifo + ":DMT-STRW_beta_im_raw").c_str(), betaImTS );
		serveData( (ifo + ":DMT-STRW_alphabeta_re_raw").c_str(), alphabetaReTS );
		serveData( (ifo + ":DMT-STRW_alphabeta_im_raw").c_str(), alphabetaImTS );
/*		
		serveData( (ifo + ":DMT-STWR_alpha_re_expave").c_str(), araveTS );
		serveData( (ifo + ":DMT-STWR_alpha_im_expave").c_str(), aiaveTS );
		serveData( (ifo + ":DMT-STWR_beta_re_expave").c_str(), braveTS );
		serveData( (ifo + ":DMT-STWR_beta_im_expave").c_str(), biaveTS );
		serveData( (ifo + ":DMT-STWR_alphabeta_re_expave").c_str(), abraveTS );
		serveData( (ifo + ":DMT-STWR_alphabeta_im_expave").c_str(), abiaveTS );
*/
		serveData( (ifo + ":DMT-STRW_alpha_re_spectrum").c_str(), &arFS );
		serveData( (ifo + ":DMT-STRW_alpha_im_spectrum").c_str(), &aiFS );
		serveData( (ifo + ":DMT-STRW_beta_re_spectrum").c_str(), &brFS );
		serveData( (ifo + ":DMT-STRW_beta_im_spectrum").c_str(), &biFS );
		serveData( (ifo + ":DMT-STRW_alphabeta_re_spectrum").c_str(), &abrFS );
		serveData( (ifo + ":DMT-STRW_alphabeta_im_spectrum").c_str(), &abiFS );
		
		serveData( (ifo + ":DMT-STRW_strain_spectrum").c_str(), &hFS );
		serveData( (ifo + ":DMT-STRW_residual_strain_spectrum").c_str(), &hRFS );
		serveData( (ifo + ":DMT-STRW_control_strain_spectrum").c_str(), &hCFS );

		serveData( (ifo + ":DMT-STRW_strain_exp_ave_spectrum").c_str(), &hFSave );
		serveData( (ifo + ":DMT-STRW_residual_strain_exp_ave_spectrum").c_str(), &hRFS );
		serveData( (ifo + ":DMT-STRW_control_strain_exp_ave_spectrum").c_str(), &hCFS );
		
		numSecsGoodData = 0;
		isFirstRun = false;
	} else if( curTime.totalS() > lastFillTime + timeStride ) {
		int missing_strides = (int)((curTime.totalS()-lastFillTime)/timeStride)-1;
		numSecsGoodData = 0;
		
		int junkData = (int)(missing_strides * timeStride / factorsTime +0.5);
		int i = 0;
		for( ; i < allFactorsDataLength - junkData; i++ ) {
			alphaRe[i] = alphaRe[i+junkData];
			alphaIm[i] = alphaIm[i+junkData];
			betaRe[i] = betaRe[i+junkData];
			betaIm[i] = betaIm[i+junkData];
			alphabetaRe[i] = alphabetaRe[i+junkData];
			alphabetaIm[i] = alphabetaIm[i+junkData];
/*			
			arave[i] = arave[i+junkData];
			aiave[i] = aiave[i+junkData];
			brave[i] = brave[i+junkData];
			biave[i] = biave[i+junkData];
			abrave[i] = abrave[i+junkData];
			abiave[i] = abiave[i+junkData];
*/		} for( ; i < allFactorsDataLength; i++ ) {
			alphaRe[i] = alphaIm[i] = betaRe[i] = betaIm[i]
				= alphabetaRe[i] = alphabetaIm[i] /*= arave[i]
				= aiave[i] = brave[i] = biave[i] = abrave[i]
				= abiave[i]*/ = -1;
		}
		cout << "Missed " << missing_strides << " strides :(";
	}
	
	lastFillTime = (long)curTime.totalS();
	if( !mOSC.satisfied( lockConditions) ) {
		numSecsGoodData = 0;
		
		int junkData = (int)(timeStride / factorsTime +0.5);
		int i = 0;
		for( ; i < allFactorsDataLength - junkData; i++ ) {
			alphaRe[i] = alphaRe[i+junkData];
			alphaIm[i] = alphaIm[i+junkData];
			betaRe[i] = betaRe[i+junkData];
			betaIm[i] = betaIm[i+junkData];
			alphabetaRe[i] = alphabetaRe[i+junkData];
			alphabetaIm[i] = alphabetaIm[i+junkData];
/*			
			arave[i] = arave[i+junkData];
			aiave[i] = aiave[i+junkData];
			brave[i] = brave[i+junkData];
			biave[i] = biave[i+junkData];
			abrave[i] = abrave[i+junkData];
			abiave[i] = abiave[i+junkData];
*/		} for( ; i < allFactorsDataLength; i++ ) {
			alphaRe[i] = alphaIm[i] = betaRe[i] = betaIm[i]
				= alphabetaRe[i] = alphabetaIm[i] /*= arave[i]
				= aiave[i] = brave[i] = biave[i] = abrave[i]
				= abiave[i]*/ = 0;
		}
		
		alphaReTS->setData( historyStartTime, factorsTime, alphaRe, allFactorsDataLength );
		alphaImTS->setData( historyStartTime, factorsTime, alphaIm, allFactorsDataLength );
		betaReTS->setData( historyStartTime, factorsTime, betaRe, allFactorsDataLength );
		betaImTS->setData( historyStartTime, factorsTime, betaIm, allFactorsDataLength );
		alphabetaReTS->setData( historyStartTime, factorsTime, alphabetaRe, allFactorsDataLength );
		alphabetaImTS->setData( historyStartTime, factorsTime, alphabetaIm, allFactorsDataLength );
/*		
		araveTS->setData( historyStartTime, factorsTime, arave, allFactorsDataLength );
		aiaveTS->setData( historyStartTime, factorsTime, aiave, allFactorsDataLength );
		braveTS->setData( historyStartTime, factorsTime, brave, allFactorsDataLength );
		biaveTS->setData( historyStartTime, factorsTime, biave, allFactorsDataLength );
		abraveTS->setData( historyStartTime, factorsTime, abrave, allFactorsDataLength );
		abiaveTS->setData( historyStartTime, factorsTime, abiave, allFactorsDataLength );
*/		
		cout << "Data not good." << endl;
		return;
	}
	
	cout << "taking data" << endl;
	
	TSeries *asq = getDacc().refData( asqChannel.c_str() );
	TSeries *darm = getDacc().refData( darmChannel.c_str() );
	TSeries *darmErr = getDacc().refData( darmErrChannel.c_str() );
	TSeries *exc = getDacc().refData( excChannel.c_str() );
	
	const int length = asq->getNSample();

	int insertAt = (int)(wings / asq->getTStep() +0.5);
	if( numSecsGoodData == 0 ) segmentStartTime = (long)(asq->getStartTime().totalS());
	if( numSecsGoodData < wings ) {
		insertAt = (int)(numSecsGoodData / asq->getTStep() +0.5);
		cout << "Not enough data. Copying data for later. Inserting at " << insertAt << endl;

		asq->getData( length, &asqData[insertAt]	);
		darm->getData( length, &darmData[insertAt] );
		darmErr->getData( length, &darmErrData[insertAt] );
		exc->getData( length, &excData[insertAt] );
		
		numSecsGoodData+=timeStride;
		cout << "new data: asq(0) = " << asqData[insertAt] << endl;
		return;
	} else if( numSecsGoodData > wings ) {
		cout << "Too much data. Moving data down." << endl;
		segmentStartTime += timeStride;

		for( int i = 0; i < insertAt; i++ ) {
			asqData[i] = asqData[i+length];
			darmData[i] = darmData[i+length];
			darmErrData[i] = darmErrData[i+length];
			excData[i] = excData[i+length];
		}
	}

	cout << "copying data - length: " << length << endl;
	cout << "Inserting at: " << insertAt << endl;

	asq->getData( length, &asqData[insertAt] );
	darm->getData( length, &darmData[insertAt] );
	darmErr->getData( length, &darmErrData[insertAt] );
	exc->getData( length, &excData[insertAt] );

	for( int i = 0; i <= wings; i++ ) {
		cout << "asq(" << i << "): " << asqData[(int)(i/asq->getTStep() +0.5)] << endl;
	}

	numSecsGoodData+=timeStride;

	cout << "sending data" << endl;
	
	double *hC, *h, *hR, *alpha, *beta, *alphabeta;
	


	strGet->computeStrain( segmentStartTime, asq->getTStep(), length * (wings+1),
		asqData, darmData, darmErrData, excData, &h, &hC, &hR,
		&alpha, &beta, &alphabeta );
	
	cout << "h(25) at receiving end: " << h[25] << endl;
	cout << "alphaRe(1) at receiving: " << alpha[2] << endl;
	
	for( int i = 0; i < allFactorsDataLength - factorDataLength; i++ ) {
		alphaRe[i] = alphaRe[i+factorDataLength];
		alphaIm[i] = alphaIm[i+factorDataLength];
		betaRe[i] = betaRe[i+factorDataLength];
		betaIm[i] = betaIm[i+factorDataLength];
		alphabetaRe[i] = alphabetaRe[i+factorDataLength];
		alphabetaIm[i] = alphabetaIm[i+factorDataLength];
/*		
		arave[i] = arave[i+factorDataLength];
		aiave[i] = arave[i+factorDataLength];
		brave[i] = arave[i+factorDataLength];
		biave[i] = arave[i+factorDataLength];
		abrave[i] = arave[i+factorDataLength];
		abiave[i] = arave[i+factorDataLength];
*/
	}
	
	for( int i = 0; i < (int)((wings+1)/factorsTime +0.5); i++ ) {
		int pos = allFactorsDataLength - factorDataLength + i;
		alphaRe[pos] = alpha[2*i];
		alphaIm[pos] = alpha[2*i+1];
		betaRe[pos] = beta[2*i];
		betaIm[pos] = beta[2*i+1];
		alphabetaRe[pos] = alphabeta[2*i];
		alphabetaIm[pos] = alphabeta[2*i+1];
/*		
		arave[pos] = ((arave[pos-1] > 0) ? 
			(EXP_FACTOR * arave[pos-1] + (1 - EXP_FACTOR) * alpha[2*i]) : alpha[2*i]);
		aiave[pos] = ((aiave[pos-1] > 0) ? 
			(EXP_FACTOR * aiave[pos-1] + (1 - EXP_FACTOR) * alpha[2*i+1]) : alpha[2*i+1]);
		brave[pos] = ((brave[pos-1] > 0) ? 
			(EXP_FACTOR * brave[pos-1] + (1 - EXP_FACTOR) * beta[2*i]) : beta[2*i]);
		biave[pos] = ((biave[pos-1] > 0) ? 
			(EXP_FACTOR * biave[pos-1] + (1 - EXP_FACTOR) * beta[2*i+1]) : beta[2*i+1]);
		abrave[pos] = ((abrave[pos-1] > 0) ? 
			(EXP_FACTOR * abrave[pos-1] + (1 - EXP_FACTOR) * alphabeta[2*i]) : alphabeta[2*i]);
		abiave[pos] = ((abiave[pos-1] > 0) ? 
			(EXP_FACTOR * abiave[pos-1] + (1 - EXP_FACTOR) * alphabeta[2*i+1]) : alphabeta[2*i+1]);
*/	}
	
	alphaReTS->setData( historyStartTime, factorsTime, alphaRe, allFactorsDataLength );
	alphaImTS->setData( historyStartTime, factorsTime, alphaIm, allFactorsDataLength );
	betaReTS->setData( historyStartTime, factorsTime, betaRe, allFactorsDataLength );
	betaImTS->setData( historyStartTime, factorsTime, betaIm, allFactorsDataLength );
	alphabetaReTS->setData( historyStartTime, factorsTime, alphabetaRe, allFactorsDataLength );
	alphabetaImTS->setData( historyStartTime, factorsTime, alphabetaIm, allFactorsDataLength );
/*	
	araveTS->setData( historyStartTime, factorsTime, arave, allFactorsDataLength );
	aiaveTS->setData( historyStartTime, factorsTime, aiave, allFactorsDataLength );
	braveTS->setData( historyStartTime, factorsTime, brave, allFactorsDataLength );
	biaveTS->setData( historyStartTime, factorsTime, biave, allFactorsDataLength );
	abraveTS->setData( historyStartTime, factorsTime, abrave, allFactorsDataLength );
	abiaveTS->setData( historyStartTime, factorsTime, abiave, allFactorsDataLength );
*/	
	Time t0 = asq->getStartTime();
	const TSeries arTS( t0, factorsTime, (int)(timeStride / factorsTime +0.5), 
		&alphaRe[allFactorsDataLength - factorDataLength] );
	const TSeries aiTS( t0, factorsTime, (int)(timeStride / factorsTime +0.5), 
		&alphaIm[allFactorsDataLength - factorDataLength] );
	const TSeries brTS( t0, factorsTime, (int)(timeStride / factorsTime +0.5), 
		&betaRe[allFactorsDataLength - factorDataLength] );
	const TSeries biTS( t0, factorsTime, (int)(timeStride / factorsTime +0.5), 
		&betaIm[allFactorsDataLength - factorDataLength] );
	const TSeries abrTS( t0, factorsTime, (int)(timeStride / factorsTime +0.5), 
		&alphabetaRe[allFactorsDataLength - factorDataLength] );
	const TSeries abiTS( t0, factorsTime, (int)(timeStride / factorsTime +0.5), 
		&alphabetaIm[allFactorsDataLength - factorDataLength] );

	psd->generate( arFS, &arTS );
	psd->generate( aiFS, &aiTS );
	psd->generate( brFS, &brTS );
	psd->generate( biFS, &biTS );
	psd->generate( abrFS, &abrTS );
	psd->generate( abiFS, &abiTS );
	
	//Scale factor to avoid underflow
	for( int i = 0; i < length; i++ ) {
		h[i+length*wings/2]*=100;
		hR[i+length*wings/2]*=100;
		hC[i+length*wings/2]*=100;
	}
	
	const TSeries hTS( t0, asq->getTStep(), length, &h[length*wings/2] );
	const TSeries hRTS( t0, asq->getTStep(), length, &hR[length*wings/2] );
	const TSeries hCTS( t0, asq->getTStep(), length, &hC[length*wings/2] );
/*	
	hRTS1min->Append( hRTS,1 );
	if( int(curTime.totalS()) % 60 == 0 ) {
		Interval i = hRTS1min->getInterval();
		if( i.GetS() == 60 ) {
			psd1min->generate( hRFS, hRTS1min );
			float *rhVals = new float[hRFS.getNStep() +1];
			hRFS.getData( hRFS.getNStep()+1, rhVals );
			for( unsigned int i = 0; i <= hFS.getNStep(); i++ ) {
				rhVals[i] = (float)sqrt( rhVals[i] )/100;
			}
			hRFS.setData(hRFS.getNStep()+1, rhVals);
			delete rhVals;
		}
		delete hRTS1min;
		hRTS1min = new TSeries( curTime, factorsTime, 0, fake );
	}
*/
	

	psd->generate( hFS, &hTS );
	psd->generate( hRFS, &hRTS );
	psd->generate( hCFS, &hCTS );

	float *hVals = new float[hFS.getNStep() + 1];
	float *rhVals = new float[hRFS.getNStep() + 1];
	float *chVals = new float[hCFS.getNStep() + 1];
	
	hFS.getData( hFS.getNStep()+1, hVals );
	hRFS.getData( hRFS.getNStep()+1, rhVals );
	hCFS.getData( hCFS.getNStep()+1, chVals );
	
	for( unsigned int i = 0; i < hFS.getNStep()+1; i++ ) {
		//why is it 100 and not 10?!!!
		//gives same vals as SenseMon, but sqrt(100)=10...
		hVals[i] = (float)sqrt( hVals[i] )/100;
		rhVals[i] = (float)sqrt( rhVals[i] )/100;
		chVals[i] = (float)sqrt( chVals[i] )/100;
	}
	
	hFS.setData( hFS.getNStep()+1, hVals );
	hRFS.setData( hRFS.getNStep()+1, rhVals );
	hCFS.setData( hCFS.getNStep()+1, chVals );

	if( !aveDataSet ) {
		hFSave.setData( hFS.getNStep()+1, hVals );
		hRFSave.setData( hRFS.getNStep()+1, rhVals );
		hCFSave.setData( hCFS.getNStep()+1, chVals );
		aveDataSet = true;
	} else {
		float *hAveVals = new float[hFSave.getNStep()+1];
		float *rhAveVals = new float[hRFSave.getNStep()+1];
		float *chAveVals = new float[hCFSave.getNStep()+1];
		
		for( unsigned int i = 0; i < hFSave.getNStep()+1; i++ ) {
			double lambda = .5;
			hAveVals[i] = (1.-lambda)*hAveVals[i] + lambda*hVals[i];
			rhAveVals[i] = (1.-lambda)*rhAveVals[i] + lambda*rhVals[i];
			chAveVals[i] = (1.-lambda)*chAveVals[i] + lambda*chVals[i];
		}
		
		hFSave.setData( hFSave.getNStep()+1, hAveVals );
		hRFSave.setData( hRFSave.getNStep()+1, rhAveVals );
		hCFSave.setData( hCFSave.getNStep()+1, chAveVals );
		
		delete hAveVals;
		delete rhAveVals;
		delete chAveVals;
	}
	
	delete hVals;
	delete rhVals;
	delete chVals;
	
	cout << "done processing data" << endl << endl;
}

void StrainWatch::Attention() {
	MonServer::Attention();
}

StrainWatch::~StrainWatch() {
	cout << "Terminating." << endl;
	
	delete strGet;

	delete [] asqData;
	delete [] darmData;
	delete [] darmErrData;
	delete [] excData;
	
	delete [] alphaRe;
	delete [] alphaIm;
	delete [] betaRe;
	delete [] betaIm;
	delete [] alphabetaRe;
	delete [] alphabetaIm;
	
	delete alphaReTS;
	delete alphaImTS;
	delete betaReTS;
	delete betaImTS;
	delete alphabetaReTS;
	delete alphabetaImTS;
	
//	delete hRTS1min;
/*	
	delete araveTS;
	delete aiaveTS;
	delete braveTS;
	delete biaveTS;
	delete abraveTS;
	delete abiaveTS;
*/	
	delete window;
	delete psd;
	
}
