#ifndef IIR_DESIGN_HH
#define IIR_DESIGN_HH

#include "IIRFilter.hh"

/** @name IIR Filter Design
  * @memo Filter design functions
  */
//@{

/** Make an butterworth filter.
  * Throws a runtime_error or invalid_argument exception, if an illegale 
  * argument is specified.
  * @memo Make an butterworth filter.
  * @param type Filter type 0: Low Pass 2: High Pass 3: Band Pass 4:Band Stop
  * @param order Filter order.
  * @param fs Sampling frequency (Hz).
  * @param f1 Pass band edge (Hz).
  * @param f2 Another pass band edge (Hz).
  * @param prewarp True for prewarping freqencies.
  * @return IIRFilter
  */
   IIRFilter butter ( Filter_Type type, int order, 
                    double fs, double f1, double f2 = 0.0, bool prewarp = true );

/** Make an chebyshev filter of type 1.
  * Throws a runtime_error or invalid_argument exception, if an illegale 
  * argument is specified.
  * @memo Make an chebyshev filter of type 1.
  * @param type Filter type 0: Low Pass 2: High Pass 3: Band Pass 4:Band Stop
  * @param order Filter order.
  * @param fs Sampling frequency (Hz).
  * @param f1 Pass band edge (Hz).
  * @param f2 Another pass band edge (Hz).
  * @param prewarp True for prewarping freqencies.
  * @return IIRFilter
  */
   IIRFilter cheby1 ( Filter_Type type, int order, double rp,
                    double fs, double f1, double f2 = 0.0, bool prewarp = true );

/** Make an chebyshev filter of type 2.
  * Throws a runtime_error or invalid_argument exception, if an illegale 
  * argument is specified.
  * @memo Make an chebyshev filter of type 2.
  * @param type Filter type 0: Low Pass 2: High Pass 3: Band Pass 4:Band Stop
  * @param order Filter order.
  * @param fs Sampling frequency (Hz).
  * @param f1 Pass band edge (Hz).
  * @param f2 Another pass band edge (Hz).
  * @param prewarp True for prewarping freqencies.
  * @return IIRFilter
  */
   IIRFilter cheby2 ( Filter_Type type, int order, double rs,
                    double fs, double f1, double f2 = 0.0, bool prewarp = true );

/** Make an elliptic filter.
  * Throws a runtime_error or invalid_argument exception, if an illegale 
  * argument is specified.
  * @memo Make an elliptic filter.
  * @param type Filter type 0: Low Pass 2: High Pass 3: Band Pass 4:Band Stop
  * @param order Filter order.
  * @param fs Sampling frequency (Hz).
  * @param f1 Pass band edge (Hz).
  * @param f2 Another pass band edge (Hz).
  * @param prewarp True for prewarping freqencies.
  * @return IIRFilter
  */
   IIRFilter ellip ( Filter_Type type, int order, double rp, double rs,
                   double fs, double f1, double f2 = 0.0, bool prewarp = true );

/** Make a notch filter.
  * Throws a runtime_error or invalid_argument exception, if an illegale 
  * argument is specified.
  * @memo Make a notch filter.
  * @param fs Sampling frequency
  * @param f0 Center frequency.
  * @param Q Quality factor ( Q = (Center Freq.) / (Width) ).
  * @param depth Depth (dB down) of the notch (dB).
  * @param prewarp True for prewarping freqencies.
  * @return IIRFilter
  */
   IIRFilter notch ( double fs, double f0, double Q, double depth = 0.0, 
                   bool prewarp = true );

/** Make a resonant gain filter.
  * Throws a runtime_error or invalid_argument exception, if an illegale 
  * argument is specified.
  * @memo Make a resonant gain filter.
  * @param fs Sampling frequency
  * @param f0 Center frequency.
  * @param Q Quality factor ( Q = (Center Freq.) / (Width) ).
  * @param height Height (dB up) of the peak (dB).
  * @param prewarp True for prewarping freqencies.
  * @return IIRFilter
  */
   IIRFilter resgain ( double fs, double f0, double Q, double height = 30.0, 
                     bool prewarp = true );

/** Make a comb filter.
  * Throws a runtime_error or invalid_argument exception, if an illegale 
  * argument is specified.
  * @memo Make a comb filter.
  * @param fs Sampling frequency
  * @param f0 Fundamental frequency.
  * @param Q Quality factor ( Q = (Center Freq.) / (Width) ).
  * @param amp Depth/height of notches/peaks (dB). 
  * @param N Number of harmonics.
  * @param prewarp True for prewarping freqencies.
  * @return IIRFilter
  */
   IIRFilter comb ( double fs, double f0, double Q, double amp = 0.0, int N = 0, 
                  bool prewarp = true );							

/** Make a simple pole.
  * Poles and zero can be specifed in the s-plane using units of rad/s (set 
  * plane to "s") or units of Hz ("f"). In both cases the frequency 
  * is expected to be POSITIVE. Alternativly, one can also specify normalized
  * poles and zeros ("n"). Normalized poles and zeros are expected to have
  * POSITIVE frequencies and their respective low and high frequency gains 
  * are set to 1---unless they are located a 0Hz. The default loaction is 
  * the s-plane.
  * For the s-plane the formula is $pole(f0) = 1/(s + w0)$ ["s"] or 
  * $pole(f0) = 1/(s + 2 pi f0)$ ["f"]. For a normalized pole ["n"] the 
  * formula is $pole(f0) = 1/(1 + i f/f0)$; $f0 > 0$ and 
  * $pole(f0) = 1/(i f)$; $f0 = 0$.
  *
  * Throws a runtime_error or invalid_argument exception, if an illegale 
  * argument is specified.
  *
  * @memo Make a simple pole
  * @param fs Sampling frequency
  * @param f Pole frequency.
  * @param g gain.
  * @param plane location where poles/zeros are specified
  * @param prewarp True for prewarping freqencies.
  * @return IIRFilter
  */
   IIRFilter pole ( double fs, double f0, double gain = 1.0, 
                  const char* plane = "s", bool prewarp = true );
		  
/** Make a simple zero. 
  * Poles and zero can be specifed in the s-plane using units of rad/s (set 
  * plane to "s") or units of Hz ("f"). In both cases the frequency 
  * is expected to be POSITIVE. Alternativly, one can also specify normalized
  * poles and zeros ("n"). Normalized poles and zeros are expected to have
  * POSITIVE frequencies and their respective low and high frequency gains 
  * are set to 1---unless they are located a 0Hz. The default loaction is 
  * the s-plane.
  * For the s-plane the formula is zero(f0) = (s + w0)$ ["s"] or 
  * $zero(f0) = (s + 2 pi f0)$ ["f"]. For a normalized zero ["n"] the 
  * formula is $zero(f0) = (1 + i f/f0)$; $f0 > 0$ and 
  * $zero(f0) = (i f)$; $f0 = 0$.
  *
  * Throws a runtime_error or invalid_argument exception, if an illegale 
  * argument is specified.
  *
  * @memo Make a simple zero
  * @param fs Sampling frequency
  * @param f Zero frequency.
  * @param g gain.
  * @param plane location where poles/zeros are specified
  * @param prewarp True for prewarping freqencies.
  * @return IIRFilter
  */
   IIRFilter zero ( double fs, double f0, double gain = 1.0, 
                  const char* plane = "s", bool prewarp = true );

/** Make a complex pole pair. 
  * Poles and zero can be specifed in the s-plane using units of rad/s (set 
  * plane to "s") or units of Hz ("f"). In both cases the frequency 
  * is expected to be POSITIVE. Alternativly, one can also specify normalized
  * poles and zeros ("n"). Normalized poles and zeros are expected to have
  * POSITIVE frequencies and their respective low and high frequency gains 
  * are set to 1---unless they are located a 0Hz. The default loaction is 
  * the s-plane.
  * For the s-plane the formula is $pole2(f0,Q)=1/(s^2 + w0 s / Q + w0^2)$ 
  * ["s"] or $pole2(f0,Q)=1/(s^2 + 2 pi f0 s / Q + (2 pi f0)^2)$ ["f"]. For 
  * a normalized complex pole pair ["n"] the  formula is 
  * $pole2(f0) = 1/(1 + i f/f0 + (f/f0)^2)$ with $f0 > 0$.
  *
  * Throws a runtime_error or invalid_argument exception, if an illegale 
  * argument is specified.
  *
  * @memo Make a complex pole pair
  * @param fs Sampling frequency
  * @param f Pole frequency.
  * @param Q quality factor
  * @param g gain
  * @param plane location where poles/zeros are specified
  * @param prewarp True for prewarping freqencies.
  * @return IIRFilter
  */
   IIRFilter pole2 ( double fs, double f0, double Q, double gain = 1.0, 
                   const char* plane = "s", bool prewarp = true );

/** Make a complex zero pair.
  * Poles and zero can be specifed in the s-plane using units of rad/s (set 
  * plane to "s") or units of Hz ("f"). In both cases the frequency 
  * is expected to be POSITIVE. Alternativly, one can also specify normalized
  * poles and zeros ("n"). Normalized poles and zeros are expected to have
  * POSITIVE frequencies and their respective low and high frequency gains 
  * are set to 1---unless they are located a 0Hz. The default loaction is 
  * the s-plane.
  * For the s-plane the formula is $zero2(f0,Q)=(s^2 + w0 s / Q + w0^2)$ 
  * ["s"] or $zero2(f0,Q)=(s^2 + 2 \pi f0 s / Q + (2 \pi f0)^2)$ ["f"]. For 
  * a normalized complex pole pair ["n"] the  formula is 
  * $pole2(f0) = (1 + i f/f0 + (f/f0)^2)$ with $f0 > 0$.
  *
  * Throws a runtime_error or invalid_argument exception, if an illegale 
  * argument is specified.
  *
  * @memo Make a complex zero pair
  * @param fs Sampling frequency
  * @param f Zero frequency.
  * @param Q quality factor
  * @param g gain.
  * @param plane location where poles/zeros are specified
  * @param prewarp True for prewarping freqencies.
  * @return IIRFilter
  */
   IIRFilter zero2 ( double fs, double f0, double Q, double gain = 1.0, 
                   const char* plane = "s", bool prewarp = true );

/** Make an IIR filter from zeros and poles.
  * Poles and zero can be specifed in the s-plane using units of rad/s (set 
  * plane to "s") or units of Hz ("f"). In both cases the real part of
  * the roots are expected to be NEAGTIVE. Alternativly, one can also specify 
  * normalized poles and zeros ("n"). Normalized poles and zeros are expected 
  * to have POSITIVE real parts and their respective low and high frequency 
  * gains are set to 1---unless they are located a 0Hz. The default loaction 
  * is the s-plane.
  * For the s-plane the formula is
  * $$zpk(s) = k \frac{(s - s_1)(s - s_2) \cdots}{(s - p_1)(s - p_2) \cdots}$$
  * with $s_1$, $s_2$,... the loactions of the zeros and $p_1$, $p_2$,... the
  * loaction of the poles. By replacing s_1->$2 \pi fz_1$and p_1->$2 \pi fp_1$
  * one obatines the "f" representation. For normalize poles and zeros ["n"] 
  * one uses poles of the form $pole(f0) = 1/(1 + i f/f0)$; $f0 > 0$ and 
  * $pole(f0) = 1/(i f)$; $f0 = 0$. The zeros are then of the form 
  * $zero(f0) = (1 + i f/f0)$; $f0 > 0$ and $zero(f0) = (i f)$; $f0 = 0$.
  * Poles and zeros with a non-zero imaginary part must come in pairs 
  * of complex conjugates.
  *
  * Throws a runtime_error or invalid_argument exception, if an illegale 
  * argument is specified.
  *
  * @memo Make a filter from zeros, poles and gain (zpk)
  * @param fs Sampling frequency
  * @param nzeros Number of zeros
  * @param zero Array of zeros
  * @param npoles Number of poles
  * @param pole Array of poles
  * @param gain Gain
  * @param plane location where poles/zeros are specified
  * @param prewarp True for prewarping freqencies.
  * @return IIRFilter
  */
   IIRFilter zpk (double fs, int nzeros, const dComplex* zero, 
                 int npoles, const dComplex* pole, double gain = 1.0,
                 const char* plane = "s", bool prewarp = true);
   IIRFilter zpk (double fs, int nzeros, const fComplex* zero, 
                 int npoles, const fComplex* pole, double gain = 1.0,
                 const char* plane = "s", bool prewarp = true);

/** Make a filter from a rational polynomial in s. 
  * A rational polynomial in s is specified by the polynomial 
  * coefficients in the numerator and the denominator in descending 
  * order of s. Fails if for some reasons the roots can not be
  * determined.
  * 
  * The formula is
  * $$zpk(s) = k \frac{a_n s^{n_z} + a_{n-1} s^{n_z - 1} \cdots}
  *                   {b_n s^{n_p} + b_{n-1} s^{n_p - 1} \cdots}$$
  * where $a_n$, $a_{n-1}$,..., $a_0$ are the coeffciients of the 
  * polynomial in the numerator and $b_n$, $b_{n-1}$,..., $b_0$ are
  * the coefficients of the polynomial in the denominator.
  * The polynomial coefficients have to be real.
  * 
  * @memo Add a rational polynomial (rpoly)
  * @param fs Sampling frequency
  * @param nnumer Number of coefficients in numerator
  * @param numer Array of numerator coefficents
  * @param ndenom Number of coeffcients in denominator
  * @param denom Array of denominator coeffcients
  * @param gain Gain
  * @param prewarp True for prewarping freqencies.
  * @return IIRFilter
  */
   IIRFilter rpoly (double fs, int nnumer, const double* numer, 
                   int ndenom, const double* denom, double gain, 
		   bool prewarp = true);
   IIRFilter rpoly (double fs, int nnumer, const float* numer,
                   int ndenom, const float* denom, double gain, 
		   bool prewarp = true);
		 
/** Make a IIR filter from a second order section.
  *
  * Throws a runtime_error or invalid_argument exception, if an illegale 
  * argument is specified.
  *
  * @memo Make a filter from b0, b1, b2, a1 and a2
  * @param fs Sampling frequency
  * @param b0 b0
  * @param b1 b1
  * @param b2 b2
  * @param a1 a1
  * @param a2 a2
  * @return IIRFilter
  */
   IIRFilter biquad (double fs, double b0, double b1, double b2, 
                     double a1, double a2);

/** Make a IIR filter from a list of second order sections. If the format is
  * 's' (standard), the coefficents are ordered like:
  * $$ gain, b1_1, b2_1, a1_1, a2_1, b1_2, b2_2, a1_2, a2_2,... $$
  * whereas for the the format 'o' (online) the order is
  * $$ gain, a1_1, a2_1, b1_1, b2_1, a1_2, a2_2, b1_2, b2_2,... $$
  * The number of coefficients must be 4 times the number of second order
  * sections plus one.
  *
  * Throws a runtime_error or invalid_argument exception, if an illegale 
  * arguments are specified.
  *
  * @memo Make a filter from second order sections
  * @param fs Sampling frequency
  * @param nba Number of coefficients
  * @param ba List of coefficients
  * @param format Coefficient format
  * @return IIRFilter
  */
   IIRFilter sos (double fs, int nba, const double* ba, 
                  const char* format = "s");
		  
/** Make a IIR filter from the list of poles and zeros in the z-plane.
  * To be stable the z-plane poles must lie within the unity cirlce.
  *
  * Throws a runtime_error or invalid_argument exception, if an illegale 
  * arguments are specified.
  *
  * @memo Make a filter from z-plane roots
  * @param fs Sampling frequency
  * @param nzeros Number of zeros
  * @param zero Array of zeros
  * @param npoles Number of poles
  * @param pole Array of poles
  * @param gain Gain
  * @return IIRFilter
  */
   IIRFilter zroots (double fs, int nzeros, const dComplex* zero,
                     int npoles, const dComplex* pole, double gain = 1.0);
   IIRFilter zroots (double fs, int nzeros, const fComplex* zero,
                     int npoles, const fComplex* pole, double gain = 1.0);

/** Make a filter from the direct form. 
  * The direct form can be written as
  * $$H(z)=\Sum_{k=0}^{nb}b_k z^{-k} / (1-\Sum_{k=1}^{na}a_k z^{-k})$$
  * 
  * Cascaded second order sections are formed by finding the roots
  * of the direct form. The specified coefficients are $b_0$, $b_1$,...,
  * $b_{nb}$ for the numerator and $a_1$, $a_2$,..., $a_{na}$ for the
  * denominator. The argument $nb$ specifies the number of b coeffcients
  * supplied to the function minus 1, whereas $na$ is exactly the number
  * of coeffcients since $a_0$ is always 1 and omitted from the list.
  * 
  * Avoid the direct form since even fairly simple filters will run into 
  * precision problems.
  *
  * Throws a runtime_error or invalid_argument exception, if an illegale 
  * arguments are specified.
  *
  * @memo Make a filter from the direct form
  * @param fs Sampling frequency
  * @param nb Number of coefficients in numerator
  * @param b Array of b coefficents
  * @param na Number of coeffcients in denominator
  * @param a Array of denominator coeffcients exclusive a0
  * @return IIRFilter
  */
   IIRFilter direct (double fs, int nb, const double* b, 
                   int na, const double* a);
		 
/** Returns the closed loop response.
  * The returned filter is the closed loop response of the specified
  * filter. This function will compute the closed loop poles and zeros
  * from the open loop transfer function.
  *
  * Throws a runtime_error or invalid_argument exception, if an illegale 
  * arguments are specified.
  *
  * @memo Closed loop transfer function
  * @see FilterDesign.hh
  * @param filter Open loop transfer function.
  * @param k Additional gain
  * @return Closed loop transfer function.
  */
   IIRFilter closeloop (const Pipe& filter, double k);
//@}


#endif
