#ifndef WINDOW_HH
#define WINDOW_HH

#include "Pipe.hh"
#include <memory>

class DVector;

/**  The Window base class performs all the common window calculation 
  *  operations. The derived classes typically differ from one another
  *  only in the window function and the constructors needed to provide
  *  whatever variation is appropriate to the specific window type. 
  *  The Window base class forces a normalization of $\Sigma W[i]^2 = 1.0$.
  *  This maintains the power of a windowed series approximately constant, 
  *  but it overrides any normalization built into the window function.
  *
  *  The Window function calculation can save half of the computation 
  *  steps for Windows that are symmetric about the center bin ($x = \pi$). 
  *  This symmetry is assumed by default. Derived classes representing 
  *  windows that don't exhibit this symmetry should define isPiSymmetric() 
  *  to return false.
  *  @memo Windowing base class.
  *  @version 1.1 ; Modified July 21, 1999
  *  @author John G. Zweizig
  */
class Window : public Pipe {
public:
   using Pipe::apply;
   using Pipe::dataCheck;
   using Pipe::isDataValid;
  
  /// Element type of window array
  typedef double element_type;
   
  /**  Build an empty window function.
    *  @memo Default constructor.
    */
  Window(void);

  /**  Copy constructor.
    *  @memo Copy constructor.
    */
  Window (const Window& w);
  
  /**  Destroy the Window object and release the function storage.
    *  @memo Virtual destructor.
    */
  virtual ~Window();

  /**  Assignment operator.
    *  @memo Assignment operator.
    */
  Window& operator= (const Window& w);
  
  /**  The argument time series is multiplied on an element-by-element basis
    *  by the windowing function. The resulting windowed series is returned.
    *  The argument TSeries is left unchanged. The Window normalization is
    *  set to $\Sigma_{i=0}^{N-1} W[i]^2 = 1.0$, as described in the class
    *  description.
    *  @memo Return a windowed TSeries.
    */
  TSeries apply(const TSeries& ts);

  /**  Clone a window function. The clone method returns a pointer to a new
    *  Window object identical to the current one.
    */
  virtual Window* clone(void) const = 0;

  /**  Check the data for validity. If the data are not applicable for 
    *  this window, an exception is thrown.
    */
  void dataCheck(const TSeries& ts) const;

  /**  Check the data for validity. Performs the same data checks as 
    *  dataCheck() but returns a boolean status instead of throwing an
    *  exception.
    */
  bool isDataValid(const TSeries& ts) const;

  /**  Return the current window length.
    *  @memo Ge the current window length.
    */
  int getLength(void) const;

  /**  Return the function RMS.
    *  @memo Get window function RMS.
    */
  double getRMS(void);

  /**  The Window table length is set to the specified value. A new function 
    *  array is allocated if insufficient storage is available. The function 
    *  array is then filled with the calculated window coefficients.
    *  @memo Set the length and calculate the Window function.
    *  @param N Number of entries in the current window function table.
    */
  void setWindow(int N);

  /**  The window function is calculated at the argument position 
    *  ($arg = 2 \pi n/(N-1)$, n is the entry number, N is the total number 
    *  of entries)
    *  @memo Calculate the Window function.
    */
  virtual double WinFc(double arg) = 0;

  /**  Return true if the Window function is symmetric around $arg=\pi$.
    *  Functions are by default symmetric, but the symmetry assumption 
    *  may be preempted by defining isPiSymmetric() for a derived class.
    *  @memo Test the Window function for symmetry around $arg=\pi$.
    *  @return true if the window function is symmetric around $arg=\pi$.
    */
  virtual bool isPiSymmetric(void);

  /**  Get the start time of the current data epoch. The start time is 
    *  set to the time of the first filtered sample after either the
    *  filter is constructed or the filter is reset. If no data have 
    *  been windowed since the construction or reset, the start time 
    *  is given as 0.
    *  @memo Get start time.
    *  @return Current epoch start time or 0 if no data have been windowed.
    */
  Time getStartTime(void) const;

  /**  Get the current time of the current data epoch. The current time is 
    *  set with the time of the next expected sample to be filtered.
    */
  Time getCurrentTime(void) const;

  /**  Returns true if the filter is in use.
   */
  bool inUse(void) const;

  /**  Prepare the filter for a disjoint data epoch.
    *  Reset the filter. 
    */
  void reset(void);

  /**  Apply the window to a FilterIO object holding a single conainer.
    */
  FilterIO& apply(const FilterIO& in);

private:
  /**  Data vector containing the window coefficients.
    */
  std::auto_ptr<DVector> mWindow;

  /**  RMS value of window coefficients.
    */
  double mRMS;
};


#ifndef __CINT__
//======================================  Inline functions
inline double 
Window::getRMS(void) {
    return mRMS;
}

inline Time 
Window::getStartTime(void) const {
    return Time(0);
}

inline Time 
Window::getCurrentTime(void) const {
    return Time(0);
}

inline bool 
Window::inUse(void) const {
    return true;
}

inline void
Window::reset(void) {
}

inline FilterIO& 
Window::apply(const FilterIO& in) {
    return Pipe::apply (in); 
}

#endif  // !defined(__CINT__)

#endif  // WINDOW_HH
